"""
Module for system.
"""
import importlib
import inspect
import logging
from collections import OrderedDict
from typing import Dict, Optional

import numpy as np

from andes.system import System as andes_System
from andes.system import (_config_numpy, load_config_rc)
from andes.variables import FileMan

from andes.utils.misc import elapsed
from andes.utils.tab import Tab

import ams
from ams.models.group import GroupBase
from ams.routines.type import TypeBase
from ams.models import file_classes
from ams.routines import all_routines
from ams.utils.paths import get_config_path
from ams.core import Config
from ams.core.matprocessor import MatProcessor
from ams.interface import to_andes
from ams.report import Report
from ams.shared import ad_dyn_models

from ams.io.matpower import system2mpc
from ams.io.matpower import write as wrtite_m
from ams.io.xlsx import write as write_xlsx
from ams.io.json import write as write_json
from ams.io.psse import write_raw

logger = logging.getLogger(__name__)


def disable_method(func):
    def wrapper(*args, **kwargs):
        logger.warning("This method is included in ANDES but not supported in AMS.")
        return None
    return wrapper


def disable_methods(methods):
    for method in methods:
        setattr(System, method, disable_method(getattr(System, method)))


class System(andes_System):
    """
    A subclass of ``andes.system.System``, this class encapsulates data, models,
    and routines for scheduling modeling and analysis in power systems.
    Some methods  inherited from the parent class are intentionally disabled.

    Parameters
    ----------
    case : str, optional
        The path to the case file.
    name : str, optional
        Name of the system instance.
    config : dict, optional
        Configuration options for the system. Overrides the default configuration if provided.
    config_path : str, optional
        The path to the configuration file.
    default_config : bool, optional
        If True, the default configuration file is loaded.
    options : dict, optional
        Additional configuration options for the system.
    **kwargs :
        Additional configuration options passed as keyword arguments.

    Attributes
    ----------
    name : str
        Name of the system instance.
    options : dict
        A dictionary containing configuration options for the system.
    models : OrderedDict
        An ordered dictionary holding the model names and instances.
    model_aliases : OrderedDict
        An ordered dictionary holding model aliases and their corresponding instances.
    groups : OrderedDict
        An ordered dictionary holding group names and instances.
    routines : OrderedDict
        An ordered dictionary holding routine names and instances.
    types : OrderedDict
        An ordered dictionary holding type names and instances.
    mats : MatrixProcessor, None
        A matrix processor instance, initially set to None.
    mat : OrderedDict
        An ordered dictionary holding common matrices.
    exit_code : int
        Command-line exit code. 0 indicates normal execution, while other values indicate errors.
    recent : RecentSolvedRoutines, None
        An object storing recently solved routines, initially set to None.
    dyn : ANDES System, None
        linked dynamic system, initially set to None.
        It is an instance of the ANDES system, which will be automatically
        set when using ``System.to_andes()``.
    files : FileMan
        File path manager instance.
    is_setup : bool
        Internal flag indicating if the system has been set up.

    Methods
    -------
    setup:
        Set up the system.
    to_andes:
        Convert the system to an ANDES system.
    """

    def __init__(self,
                 case: Optional[str] = None,
                 name: Optional[str] = None,
                 config: Optional[Dict] = None,
                 config_path: Optional[str] = None,
                 default_config: Optional[bool] = False,
                 options: Optional[Dict] = None,
                 **kwargs
                 ):

        # TODO: might need _check_group_common
        func_to_disable = [
            # --- not sure ---
            'set_config', 'set_dae_names', 'set_output_subidx', 'set_var_arrays',
            # --- not used in AMS ---
            '_check_group_common', '_clear_adder_setter', '_e_to_dae', '_expand_pycode', '_finalize_pycode',
            '_find_stale_models', '_get_models', '_init_numba', '_load_calls', '_mp_prepare',
            '_p_restore', '_store_calls', '_store_tf', '_to_orddct', '_v_to_dae',
            'save_config', 'collect_config', 'e_clear', 'f_update',
            'fg_to_dae', 'from_ipysheet', 'g_islands', 'g_update', 'get_z',
            'init', 'j_islands', 'j_update', 'l_update_eq',
            'l_update_var', 'precompile', 'prepare', 'reload', 'remove_pycapsule',
            's_update_post', 's_update_var', 'store_adder_setter', 'store_no_check_init',
            'store_sparse_pattern', 'store_switch_times', 'switch_action', 'to_ipysheet',
            'undill']
        disable_methods(func_to_disable)

        self.name = name
        self.options = {}
        if options is not None:
            self.options.update(options)
        if kwargs:
            self.options.update(kwargs)
        self.models = OrderedDict()          # model names and instances
        self.model_aliases = OrderedDict()   # alias: model instance
        self.groups = OrderedDict()          # group names and instances
        self.routines = OrderedDict()        # routine names and instances
        self.types = OrderedDict()           # type names and instances
        self.mats = MatProcessor(self)       # matrix processor
        # TODO: there should be an exit_code for each routine
        self.exit_code = 0                   # command-line exit code, 0 - normal, others - error.
        self.recent = None                   # recent solved routines
        self.dyn = None                      # ANDES system

        # get and load default config file
        self._config_path = get_config_path()
        if config_path is not None:
            self._config_path = config_path
        if default_config is True:
            self._config_path = None

        self._config_object = load_config_rc(self._config_path)
        self._update_config_object()
        self.config = Config(self.__class__.__name__, dct=config)
        self.config.load(self._config_object)

        # custom configuration for system goes after this line
        self.config.add(OrderedDict((('freq', 60),
                                     ('mva', 100),
                                     ('seed', 'None'),
                                     ('save_stats', 0),  # TODO: not sure what this is for
                                     ('np_divide', 'warn'),
                                     ('np_invalid', 'warn'),
                                     )))

        self.config.add_extra("_help",
                              freq='base frequency [Hz]',
                              mva='system base MVA',
                              seed='seed (or None) for random number generator',
                              np_divide='treatment for division by zero',
                              np_invalid='treatment for invalid floating-point ops.',
                              )

        self.config.add_extra("_alt",
                              freq="float",
                              mva="float",
                              seed='int or None',
                              np_divide={'ignore', 'warn', 'raise', 'call', 'print', 'log'},
                              np_invalid={'ignore', 'warn', 'raise', 'call', 'print', 'log'},
                              )

        self.config.check()
        _config_numpy(seed=self.config.seed,
                      divide=self.config.np_divide,
                      invalid=self.config.np_invalid,
                      )

        # TODO: revise the following attributes, it seems that these are not used in AMS
        self._getters = dict(f=list(), g=list(), x=list(), y=list())
        self._adders = dict(f=list(), g=list(), x=list(), y=list())
        self._setters = dict(f=list(), g=list(), x=list(), y=list())

        self.files = FileMan(case=case, **self.options)    # file path manager

        # internal flags
        self.is_setup = False        # if system has been setup

        self.import_types()
        self.import_groups()
        self.import_models()
        self.import_routines()

    def import_types(self):
        """
        Import all types classes defined in ``routines/type.py``.

        Types will be stored as instances with the name as class names.
        All types will be stored to dictionary ``System.types``.
        """
        module = importlib.import_module('ams.routines.type')
        for m in inspect.getmembers(module, inspect.isclass):
            name, cls = m
            if name == 'TypeBase':
                continue
            elif not issubclass(cls, TypeBase):
                # skip other imported classes such as `OrderedDict`
                continue

            self.__dict__[name] = cls()
            self.types[name] = self.__dict__[name]

    def _collect_group_data(self, items):
        """
        Set the owner for routine attributes: `RParam`, `Var`, `ExpressionCalc`, `Expression`,
        and `RBaseService`.
        """
        for item_name, item in items.items():
            if item.model is None:
                continue
            elif item.model in self.groups.keys():
                item.is_group = True
                item.owner = self.groups[item.model]
            elif item.model in self.models.keys():
                item.owner = self.models[item.model]
            elif item.model == 'mats':
                item.owner = self.mats
            else:
                logger.debug(f'item_name: {item_name}')
                msg = f'Model indicator \'{item.model}\' of <{item.rtn.class_name}.{item_name}>'
                msg += ' is not a model or group. Likely a modeling error.'
                logger.warning(msg)

    def import_routines(self):
        """
        Import routines as defined in ``routines/__init__.py``.

        Routines will be stored as instances with the name as class names.
        All routines will be stored to dictionary ``System.routines``.

        Examples
        --------
        ``System.PFlow`` is the power flow routine instance.
        """
        for file, cls_list in all_routines.items():
            for cls_name in cls_list:
                routine = importlib.import_module('ams.routines.' + file)
                the_class = getattr(routine, cls_name)
                attr_name = cls_name
                self.__dict__[attr_name] = the_class(system=self, config=self._config_object)
                self.routines[attr_name] = self.__dict__[attr_name]
                self.routines[attr_name].config.check()
                # NOTE: the following code is not used in ANDES
                for vname, rtn in self.routines.items():
                    # TODO: collect routiens into types
                    type_name = getattr(rtn, 'type')
                    type_instance = self.types[type_name]
                    type_instance.routines[vname] = rtn
                    # self.types[rtn.type].routines[vname] = rtn
                    # Collect RParams
                    rparams = getattr(rtn, 'rparams')
                    self._collect_group_data(rparams)
                    # Collect routine Vars
                    r_vars = getattr(rtn, 'vars')
                    self._collect_group_data(r_vars)
                    # Collect ExpressionCalcs
                    exprc = getattr(rtn, 'exprcs')
                    self._collect_group_data(exprc)
                    # Collect Expressions
                    expr = getattr(rtn, 'exprs')
                    self._collect_group_data(expr)

    def import_groups(self):
        """
        Import all groups classes defined in ``models/group.py``.

        Groups will be stored as instances with the name as class names.
        All groups will be stored to dictionary ``System.groups``.
        """
        module = importlib.import_module('ams.models.group')
        for m in inspect.getmembers(module, inspect.isclass):

            name, cls = m
            if name == 'GroupBase':
                continue
            elif not issubclass(cls, GroupBase):
                # skip other imported classes such as `OrderedDict`
                continue

            self.__dict__[name] = cls()
            self.groups[name] = self.__dict__[name]

    def import_models(self):
        """
        Import and instantiate models as System member attributes.

        Models defined in ``models/__init__.py`` will be instantiated `sequentially` as attributes with the same
        name as the class name.
        In addition, all models will be stored in dictionary ``System.models`` with model names as
        keys and the corresponding instances as values.

        Examples
        --------
        ``system.Bus`` stores the `Bus` object, and ``system.PV`` stores the PV generator object.

        ``system.models['Bus']`` points the same instance as ``system.Bus``.
        """
        for fname, cls_list in file_classes:
            for model_name in cls_list:
                the_module = importlib.import_module('ams.models.' + fname)
                the_class = getattr(the_module, model_name)
                self.__dict__[model_name] = the_class(system=self, config=self._config_object)
                self.models[model_name] = self.__dict__[model_name]
                self.models[model_name].config.check()

                # link to the group
                group_name = self.__dict__[model_name].group
                self.__dict__[group_name].add_model(model_name, self.__dict__[model_name])
        # NOTE: model_aliases is not used in AMS currently
        # for key, val in ams.models.model_aliases.items():
        #     self.model_aliases[key] = self.models[val]
        #     self.__dict__[key] = self.models[val]

    def collect_ref(self):
        """
        Collect indices into `BackRef` for all models.
        """
        models_and_groups = list(self.models.values()) + list(self.groups.values())

        # create an empty list of lists for all `BackRef` instances
        for model in models_and_groups:
            for ref in model.services_ref.values():
                ref.v = [list() for _ in range(model.n)]

        # `model` is the model who stores `IdxParam`s to other models
        # `BackRef` is declared at other models specified by the `model` parameter
        # of `IdxParam`s.

        for model in models_and_groups:
            if model.n == 0:
                continue

            # skip: a group is not allowed to link to other groups
            if not hasattr(model, "idx_params"):
                continue

            for idxp in model.idx_params.values():
                if (idxp.model not in self.models) and (idxp.model not in self.groups):
                    continue
                dest = self.__dict__[idxp.model]

                if dest.n == 0:
                    continue

                for name in (model.class_name, model.group):
                    # `BackRef` not requested by the linked models or groups
                    if name not in dest.services_ref:
                        continue

                    for model_idx, dest_idx in zip(model.idx.v, idxp.v):
                        if dest_idx not in dest.uid:
                            continue

                        dest.set_backref(name,
                                         from_idx=model_idx,
                                         to_idx=dest_idx)

    def reset(self, force=False):
        """
        Reset to the state after reading data and setup.
        """
        self.is_setup = False
        self.setup()

    def add(self, model, param_dict=None, **kwargs):
        """
        Add a device instance for an existing model.

        Revised from ``andes.system.System.add()``.
        """
        if model not in self.models and (model not in self.model_aliases):
            if model in ad_dyn_models:
                logger.debug("ANDES dynamic model <%s> is skipped.", model)
            else:
                logger.warning("<%s> is not an existing model.", model)
            return

        if self.is_setup:
            raise NotImplementedError("Adding devices are not allowed after setup.")

        group_name = self.__dict__[model].group
        group = self.groups[group_name]

        if param_dict is None:
            param_dict = {}
        if kwargs is not None:
            param_dict.update(kwargs)

        # remove `uid` field
        param_dict.pop('uid', None)

        idx = param_dict.pop('idx', None)
        if idx is not None and (not isinstance(idx, str) and np.isnan(idx)):
            idx = None

        idx = group.get_next_idx(idx=idx, model_name=model)
        self.__dict__[model].add(idx=idx, **param_dict)
        group.add(idx=idx, model=self.__dict__[model])

        return idx

    def setup(self):
        """
        Set up system for studies.

        This function is to be called after adding all device data.
        """
        ret = True
        t0, _ = elapsed()

        if self.is_setup:
            logger.warning('System has been setup. Calling setup twice is not allowed.')
            ret = False
            return ret

        self.collect_ref()
        self._list2array()     # `list2array` must come before `link_ext_param`
        if not self.link_ext_param():
            ret = False

        # --- model parameters range check ---
        # TODO: there might be other parameters check?
        adjusted_params = []
        param_to_check = ['rate_a', 'rate_b', 'rate_c', 'amax', 'amin']
        for pname in param_to_check:
            param = self.Line.params[pname]
            if np.any(param.v == 0):
                adjusted_params.append(pname)
                param.v[param.v == 0] = param.default
        if adjusted_params:
            adjusted_params_str = ', '.join(adjusted_params)
            msg = f"Zero Line parameters detected, adjusted to default values: {adjusted_params_str}."
            logger.info(msg)
        # --- bus type correction ---
        pq_bus = self.PQ.bus.v
        pv_bus = self.PV.bus.v
        slack_bus = self.Slack.bus.v
        # TODO: how to include islanded buses?
        if self.Bus.n > 0 and np.all(self.Bus.type.v == 1):
            self.Bus.alter(src='type', idx=pq_bus, value=1)
            self.Bus.alter(src='type', idx=pv_bus, value=2)
            self.Bus.alter(src='type', idx=slack_bus, value=3)
            logger.info("All bus type are PQ, adjusted given load and generator connection status.")
        # === no device addition or removal after this point ===
        self.calc_pu_coeff()   # calculate parameters in system per units

        if ret is True:
            self.is_setup = True  # set `is_setup` if no error occurred
        else:
            logger.error("System setup failed. Please resolve the reported issue(s).")
            self.exit_code += 1

        a0 = 0
        for _, mdl in self.models.items():
            for _, algeb in mdl.algebs.items():
                algeb.v = np.zeros(algeb.owner.n)
                algeb.a = np.arange(a0, a0 + algeb.owner.n)
                a0 += algeb.owner.n

        # NOTE: this is a temporary solution for building Y matrix
        # consider refator this part if any other similar cases occur in the future
        self.Line.a1a = self.Bus.get(src='a', attr='a', idx=self.Line.bus1.v)
        self.Line.a2a = self.Bus.get(src='a', attr='a', idx=self.Line.bus2.v)

        # assign bus type as placeholder; 1=PQ, 2=PV, 3=ref, 4=isolated
        if self.Bus.type.v.sum() == self.Bus.n:  # if all type are PQ
            self.Bus.alter(src='type', idx=self.PV.bus.v,
                           value=np.ones(self.PV.n))
            self.Bus.alter(src='type', idx=self.Slack.bus.v,
                           value=np.ones(self.Slack.n))

        # --- assign column and row names ---
        self.mats.Cft.col_names = self.Line.idx.v
        self.mats.Cft.row_names = self.Bus.idx.v

        self.mats.CftT.col_names = self.Bus.idx.v
        self.mats.CftT.row_names = self.Line.idx.v

        self.mats.Cg.col_names = self.StaticGen.get_all_idxes()
        self.mats.Cg.row_names = self.Bus.idx.v

        self.mats.Cl.col_names = self.PQ.idx.v
        self.mats.Cl.row_names = self.Bus.idx.v

        self.mats.Csh.col_names = self.Shunt.idx.v
        self.mats.Csh.row_names = self.Bus.idx.v

        self.mats.Bbus.col_names = self.Bus.idx.v
        self.mats.Bbus.row_names = self.Bus.idx.v

        self.mats.Bf.col_names = self.Bus.idx.v
        self.mats.Bf.row_names = self.Line.idx.v

        self.mats.PTDF.col_names = self.Bus.idx.v
        self.mats.PTDF.row_names = self.Line.idx.v

        self.mats.LODF.col_names = self.Line.idx.v
        self.mats.LODF.row_names = self.Line.idx.v

        _, s = elapsed(t0)
        logger.info('System set up in %s.', s)

        return ret

    # FIXME: remove unused methods
    # # Disable methods not supported in AMS
    # func_to_include = [
    #     'import_models', 'import_groups', 'import_routines',
    #     'setup', 'init_algebs',
    #     '_update_config_object',
    #     ]
    # # disable_methods(func_to_disable)
    # __dict__ = {method: lambda self: self.x for method in func_to_include}

    def supported_routines(self, export='plain'):
        """
        Return the support type names and routine names in a table.

        Returns
        -------
        str
            A table-formatted string for the types and routines
        """

        def rst_ref(name, export):
            """
            Refer to the model in restructuredText mode so that
            it renders as a hyperlink.
            """

            if export == 'rest':
                return ":ref:`" + name + '`'
            else:
                return name

        pairs = list()
        for g in self.types:
            routines = list()
            for m in self.types[g].routines:
                routines.append(rst_ref(m, export))
            if len(routines) > 0:
                pairs.append((rst_ref(g, export), ', '.join(routines)))

        tab = Tab(title='Supported Types and Routines',
                  header=['Type', 'Routines'],
                  data=pairs,
                  export=export,
                  )

        return tab.draw()

    def connectivity(self, info=True):
        """
        Perform connectivity check for system.

        Parameters
        ----------
        info : bool
            True to log connectivity summary.
        """

        raise NotImplementedError

    def to_andes(self, addfile=None,
                 setup=False, no_output=False,
                 default_config=True,
                 verify=False, tol=1e-3,
                 **kwargs):
        """
        Convert the AMS system to an ANDES system.

        A preferred dynamic system file to be added has following features:
        1. The file contains both power flow and dynamic models.
        2. The file can run in ANDES natively.
        3. Power flow models are in the same shape as the AMS system.
        4. Dynamic models, if any, are in the same shape as the AMS system.

        This function is wrapped as the ``System`` class method ``to_andes()``.
        Using the file conversion ``to_andes()`` will automatically
        link the AMS system instance to the converted ANDES system instance
        in the AMS system attribute ``dyn``.

        It should be noted that detailed dynamic simualtion requires extra
        dynamic models to be added to the ANDES system, which can be passed
        through the ``addfile`` argument.

        Parameters
        ----------
        system : System
            The AMS system to be converted to ANDES format.
        addfile : str, optional
            The additional file to be converted to ANDES dynamic mdoels.
        setup : bool, optional
            Whether to call `setup()` after the conversion. Default is True.
        no_output : bool, optional
            To ANDES system.
        default_config : bool, optional
            To ANDES system.
        verify : bool
            If True, the converted ANDES system will be verified with the source
            AMS system using AC power flow.
        tol : float
            The tolerance of error.

        Returns
        -------
        adsys : andes.system.System
            The converted ANDES system.

        Examples
        --------
        >>> import ams
        >>> import andes
        >>> sp = ams.load(ams.get_case('ieee14/ieee14_uced.xlsx'), setup=True)
        >>> sa = sp.to_andes(addfile=andes.get_case('ieee14/ieee14_full.xlsx'),
        ...                  setup=False, overwrite=True, no_output=True)

        Notes
        -----
        1. Power flow models in the addfile will be skipped and only dynamic models will be used.
        2. The addfile format is guessed based on the file extension. Currently only ``xlsx`` is supported.
        3. Index in the addfile is automatically adjusted when necessary.
        """
        return to_andes(system=self, addfile=addfile,
                        setup=setup, no_output=no_output,
                        default_config=default_config,
                        verify=verify, tol=tol,
                        **kwargs)

    def summary(self):
        """
        Print out system summary.
        """
        # FIXME: add system connectivity check
        # logger.info("-> System connectivity check results:")
        rtn_check = OrderedDict((key, val._data_check(info=False)) for key, val in self.routines.items())
        rtn_types = OrderedDict({tp: [] for tp in self.types.keys()})

        for name, data_pass in rtn_check.items():
            if data_pass:
                r_type = self.routines[name].type
                rtn_types[r_type].append(name)

        nb = self.Bus.n
        nl = self.Line.n
        ng = self.StaticGen.n

        pd = self.PQ.p0.v.sum()
        qd = self.PQ.q0.v.sum()

        out = list()

        out.append("-> Systen size:")
        out.append(f"Base: {self.config.mva} MVA; Frequency: {self.config.freq} Hz")
        out.append(f"{nb} Buses; {nl} Lines; {ng} Static Generators")
        out.append(f"Active load: {pd:,.2f} p.u.; Reactive load: {qd:,.2f} p.u.")

        out.append("-> Data check results:")
        for rtn_type, names in rtn_types.items():
            if len(names) == 0:
                continue
            names = ", ".join(names)
            out.append(f"{rtn_type}: {names}")

        out_str = '\n'.join(out)
        logger.info(out_str)

    def report(self):
        """
        Write system routine reports to a plain-text file.

        Returns
        -------
        bool
            True if the report is written successfully.
        """
        if self.files.no_output is False:
            r = Report(self)
            r.write()
            return True

        return False

    def to_mpc(self):
        """
        Export an AMS system to a MATPOWER dict.

        Returns
        -------
        dict
            A dictionary representing the MATPOWER case.
        """
        return system2mpc(self)

    def to_m(self, outfile: str, overwrite: bool = None):
        """
        Export an AMS system to a MATPOWER M-file.

        Parameters
        ----------
        outfile : str
            The output file name.
        overwrite : bool, optional
            If True, overwrite the existing file. Default is None.
        """
        return wrtite_m(self, outfile=outfile, overwrite=overwrite)

    def to_xlsx(self, outfile: str, overwrite: bool = None):
        """
        Export an AMS system to an Excel file.

        Parameters
        ----------
        outfile : str
            The output file name.
        overwrite : bool, optional
            If True, overwrite the existing file. Default is None.
        """
        return write_xlsx(self, outfile=outfile, overwrite=overwrite)

    def to_json(self, outfile: str, overwrite: bool = None):
        """
        Export an AMS system to a JSON file.

        Parameters
        ----------
        outfile : str
            The output file name.
        overwrite : bool, optional
            If True, overwrite the existing file. Default is None.
        """
        return write_json(self, outfile=outfile, overwrite=overwrite)

    def to_raw(self, outfile: str, overwrite: bool = None):
        """
        Export an AMS system to a v33 PSS/E RAW file.

        Parameters
        ----------
        outfile : str
            The output file name.
        overwrite : bool, optional
            If True, overwrite the existing file. Default is None.
        """
        return write_raw(self, outfile=outfile, overwrite=overwrite)

# --------------- Helper Functions ---------------
# NOTE: _config_numpy, load_config_rc are imported from andes.system


def example(setup=True, no_output=True, **kwargs):
    """
    Return an :py:class:`ams.system.System` object for the
    ``ieee14_uced.xlsx`` as an example.

    This function is useful when a user wants to quickly get a
    System object for testing.

    Returns
    -------
    System
        An example :py:class:`ams.system.System` object.
    """

    return ams.load(ams.get_case('matpower/case14.m'),
                    setup=setup, no_output=no_output, **kwargs)



"""
Excel reader and writer for AMS.

This module leverages the existing parser and writer in andes.io.xlsx.
"""
import logging

from andes.io.xlsx import (read, testlines, confirm_overwrite, _add_book)  # NOQA

from ams.shared import pd, empty_adsys, ad_models


logger = logging.getLogger(__name__)


def write(system, outfile,
          skip_empty=True, overwrite=None, add_book=None,
          to_andes=False,
          ):
    """
    Write loaded AMS system data into an xlsx file

    Revised function ``andes.io.xlsx.write`` to skip non-andes models.

    Parameters
    ----------
    system : System
        A loaded system with parameters
    outfile : str
        Path to the output file
    skip_empty : bool
        Skip output of empty models (n = 0)
    overwrite : bool, optional
        None to prompt for overwrite selection; True to overwrite; False to not overwrite
    add_book : str, optional
        An optional model to be added to the output spreadsheet
    to_andes : bool, optional
        Write to an ANDES system, where non-ANDES models are skipped

    Returns
    -------
    bool
        True if file written; False otherwise
    """
    if not confirm_overwrite(outfile, overwrite=overwrite):
        return False

    writer = pd.ExcelWriter(outfile, engine='xlsxwriter')
    writer = _write_system(system, writer, skip_empty, to_andes=to_andes)
    writer = _add_book(system, writer, add_book)

    writer.close()

    logger.info('xlsx file written to "%s"', outfile)
    return True


def _write_system(system, writer, skip_empty, to_andes=False):
    """
    Write the system to pandas ExcelWriter

    Rewrite function ``andes.io.xlsx._write_system`` to skip non-andes sheets.
    """
    for name, instance in system.models.items():
        if skip_empty and instance.n == 0:
            continue
        instance.cache.refresh("df_in")
        if to_andes:
            if name not in ad_models:
                continue
            # NOTE: ommit parameters that are not in ANDES
            skip_params = []
            ams_params = list(instance.params.keys())
            andes_params = list(empty_adsys.models[name].params.keys())
            skip_params = list(set(ams_params) - set(andes_params))
            df = instance.cache.df_in.drop(skip_params, axis=1, errors='ignore')
        else:
            df = instance.cache.df_in
        df.to_excel(writer, sheet_name=name, freeze_panes=(1, 0))
    return writer



"""
Json reader and writer for AMS.

This module leverages the existing parser and writer in andes.io.json.
"""
import json
import logging

from collections import OrderedDict

from andes.io.json import (testlines, read)  # NOQA
from andes.utils.paths import confirm_overwrite

from ams.shared import empty_adsys, ad_models

logger = logging.getLogger(__name__)


def write(system, outfile, skip_empty=True, overwrite=None,
          to_andes=False,
          ):
    """
    Write loaded AMS system data into an json file.
    If to_andes is True, only write models that are in ANDES,
    but the outfile might not be able to be read back into AMS.

    Revise function ``andes.io.json.write`` to skip non-andes models.

    Parameters
    ----------
    system : System
        A loaded system with parameters
    outfile : str
        Path to the output file
    skip_empty : bool
        Skip output of empty models (n = 0)
    overwrite : bool, optional
        None to prompt for overwrite selection; True to overwrite; False to not overwrite
    to_andes : bool, optional
        Write to an ANDES system, where non-ANDES models are skipped

    Returns
    -------
    bool
        True if file written; False otherwise
    """
    if not confirm_overwrite(outfile, overwrite):
        return False

    if hasattr(outfile, 'write'):
        outfile.write(_dump_system(system, skip_empty, to_andes=to_andes))
    else:
        with open(outfile, 'w') as writer:
            writer.write(_dump_system(system, skip_empty, to_andes=to_andes))
            logger.info('JSON file written to "%s"', outfile)

    return True


def _dump_system(system, skip_empty, orient='records', to_andes=False):
    """
    Dump parameters of each model into a json string and return
    them all in an OrderedDict.
    """
    out = OrderedDict()
    for name, instance in system.models.items():
        if skip_empty and instance.n == 0:
            continue
        if to_andes:
            if name not in ad_models:
                continue
            # NOTE: ommit parameters that are not in ANDES
            skip_params = []
            ams_params = list(instance.params.keys())
            andes_params = list(empty_adsys.models[name].params.keys())
            skip_params = list(set(ams_params) - set(andes_params))
            df = instance.cache.df_in.drop(skip_params, axis=1, errors='ignore')
            out[name] = df.to_dict(orient=orient)
        else:
            df = instance.cache.df_in
            out[name] = df.to_dict(orient=orient)
    return json.dumps(out, indent=2)



"""
Module for optimization Var.
"""
import logging

from typing import Optional, Union
from collections import OrderedDict
import re

import numpy as np

from andes.core.common import Config

import cvxpy as cp

from ams.utils import pretty_long_message
from ams.shared import _prefix, _max_length

from ams.opt import OptzBase, ensure_symbols, ensure_mats_and_parsed

logger = logging.getLogger(__name__)


class Var(OptzBase):
    """
    Base class for variables used in a routine.

    When `horizon` is provided, the variable will be expanded to a matrix,
    where rows are indexed by the source variable index and columns are
    indexed by the horizon index.

    Parameters
    ----------
    info : str, optional
        Descriptive information
    unit : str, optional
        Unit
    tex_name : str
        LaTeX-formatted variable symbol. Defaults to the value of ``name``.
    name : str, optional
        Variable name. One should typically assigning the name directly because
        it will be automatically assigned by the model. The value of ``name``
        will be the symbol name to be used in expressions.
    src : str, optional
        Source variable name. Defaults to the value of ``name``.
    model : str, optional
        Name of the owner model or group.
    horizon : ams.routines.RParam, optional
        Horizon idx.
    nonneg : bool, optional
        Non-negative variable
    nonpos : bool, optional
        Non-positive variable
    cplx : bool, optional
        Complex variable
    imag : bool, optional
        Imaginary variable
    symmetric : bool, optional
        Symmetric variable
    diag : bool, optional
        Diagonal variable
    psd : bool, optional
        Positive semi-definite variable
    nsd : bool, optional
        Negative semi-definite variable
    hermitian : bool, optional
        Hermitian variable
    boolean : bool, optional
        Boolean variable
    integer : bool, optional
        Integer variable
    pos : bool, optional
        Positive variable
    neg : bool, optional
        Negative variable

    Attributes
    ----------
    a : np.ndarray
        Variable address.
    _v : np.ndarray
        Local-storage of the variable value.
    rtn : ams.routines.Routine
        The owner routine instance.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 src: Optional[str] = None,
                 unit: Optional[str] = None,
                 model: Optional[str] = None,
                 shape: Optional[Union[tuple, int]] = None,
                 v0: Optional[str] = None,
                 horizon: Optional[str] = None,
                 nonneg: Optional[bool] = False,
                 nonpos: Optional[bool] = False,
                 cplx: Optional[bool] = False,
                 imag: Optional[bool] = False,
                 symmetric: Optional[bool] = False,
                 diag: Optional[bool] = False,
                 psd: Optional[bool] = False,
                 nsd: Optional[bool] = False,
                 hermitian: Optional[bool] = False,
                 boolean: Optional[bool] = False,
                 integer: Optional[bool] = False,
                 pos: Optional[bool] = False,
                 neg: Optional[bool] = False,
                 ):
        self.name = name
        self.info = info
        self.unit = unit

        self.tex_name = tex_name if tex_name else name
        # variable internal index inside a model (assigned in run time)
        self.id = None
        OptzBase.__init__(self, name=name, info=info, unit=unit, model=model)
        self.src = src
        self.v0 = v0
        self.horizon = horizon
        self._shape = shape
        self._v = None
        self.a: np.ndarray = np.array([], dtype=int)

        self.config = Config(name=self.class_name)  # `config` that can be exported

        self.config.add(OrderedDict((('nonneg', nonneg),
                                     ('nonpos', nonpos),
                                     ('complex', cplx),
                                     ('imag', imag),
                                     ('symmetric', symmetric),
                                     ('diag', diag),
                                     ('psd', psd),
                                     ('nsd', nsd),
                                     ('hermitian', hermitian),
                                     ('boolean', boolean),
                                     ('integer', integer),
                                     ('pos', pos),
                                     ('neg', neg),
                                     )))

    @property
    def v(self):
        """
        Return the CVXPY variable value.
        """
        if self.optz is None:
            return None
        if self.optz.value is None:
            try:
                shape = self.optz.shape
                return np.zeros(shape)
            except AttributeError:
                return None
        else:
            return self.optz.value

    @v.setter
    def v(self, value):
        if self.optz is None:
            logger.info(f"Variable <{self.name}> is not initialized yet.")
        else:
            self.optz.value = value

    @ensure_symbols
    def parse(self):
        """
        Parse the variable.
        """
        sub_map = self.om.rtn.syms.sub_map
        # NOTE: number of rows is the size of the source variable
        if self.owner is not None:
            nr = self.owner.n
            nc = 0
            if self.horizon:
                # NOTE: numer of columns is the horizon if exists
                nc = int(self.horizon.n)
                shape = (nr, nc)
            else:
                shape = (nr,)
        elif isinstance(self._shape, int):
            shape = (self._shape,)
            nr = shape
            nc = 0
        elif isinstance(self._shape, tuple):
            shape = self._shape
            nr = shape[0]
            nc = shape[1] if len(shape) > 1 else 0
        else:
            raise ValueError(f"Invalid shape {self._shape}.")
        code_var = f"var({shape}, **config)"
        logger.debug(f" - Var <{self.name}>: {self.code}")
        for pattern, replacement, in sub_map.items():
            try:
                code_var = re.sub(pattern, replacement, code_var)
            except Exception as e:
                raise Exception(f"Error in parsing var <{self.name}>.\n{e}")
        self.code = code_var
        return True

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the variable.

        Returns
        -------
        bool
            Returns True if the evaluation is successful, False otherwise.
        """
        # NOTE: in CVXPY, Config only allow lower case letters
        config = {}  # used in `self.code`
        for k, v in self.config.as_dict().items():
            if k == 'psd':
                config['PSD'] = v
            elif k == 'nsd':
                config['NSD'] = v
            elif k == 'bool':
                config['boolean'] = v
            else:
                config[k] = v
        msg = f" - Var <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        try:
            local_vars = {'self': self, 'config': config, 'cp': cp}
            self.optz = eval(self.code, {}, local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating Var <{self.name}>.\n{e}")
        return True

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.owner.__class__.__name__}.{self.name}'



"""
Model for rolling horizon used in scheduling.
"""

from andes.core import ModelData, NumParam
from andes.models.timeseries import str_list_iconv
from ams.core.model import Model


def str_list_oconv(x):
    """
    Convert list into a list literal.
    """
    # NOTE: convert elements to string from number first, then join them
    str_x = [str(i) for i in x]
    return ','.join(str_x)


class TimeSlot(ModelData, Model):
    """
    Base model for time slot data used in multi-interval scheduling.
    """

    def __init__(self, system=None, config=None):
        ModelData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Horizon'

        self.sd = NumParam(info='zonal load scaling factor',
                           tex_name=r's_{d}',
                           iconvert=str_list_iconv,
                           oconvert=str_list_oconv,
                           vtype=float)


class EDTSlot(TimeSlot):
    """
    Time slot model for ED.

    `sd` is the zonal load scaling factor.
    Cells in `sd` should have `nz` values seperated by comma,
    where `nz` is the number of `Zone` in the system.

    `ug` is the unit commitment decisions.
    Cells in `ug` should have `ng` values seperated by comma,
    where `ng` is the number of `StaticGen` in the system.
    """

    def __init__(self, system=None, config=None):
        TimeSlot.__init__(self, system, config)

        self.ug = NumParam(info='unit commitment decisions',
                           tex_name=r'u_{g}',
                           iconvert=str_list_iconv,
                           oconvert=str_list_oconv,
                           vtype=int)


class UCTSlot(TimeSlot):
    """
    Time slot model for UC.

    `sd` is the zonal load scaling factor.
    Cells in `sd` should have `nz` values seperated by comma,
    where `nz` is the number of `Zone` in the system.
    """

    def __init__(self, system=None, config=None):
        TimeSlot.__init__(self, system, config)



"""
PSS/E .raw reader for AMS.
This module is the existing module in ``andes.io.psse``.
"""

from andes.io.psse import testlines  # NOQA
from andes.io.psse import read as ad_read
from andes.io.xlsx import confirm_overwrite
from andes.shared import rad2deg, pd

from ams import __version__ as version
from ams.shared import copyright_msg, nowarranty_msg, report_time


def read(system, file):
    """
    Read PSS/E RAW file v32/v33 formats.

    Revised from ``andes.io.psse.read`` to complete model ``Zone`` when necessary.
    """
    ret = ad_read(system, file)
    # Extract zone data
    zone = system.Bus.zone.v
    zone_map = {}

    # Check if there are zones to process
    # NOTE: since `Zone` and `Area` below to group `Collection`, we add
    # numerical Zone idx after the last Area idx.
    if zone:
        n_zone = system.Area.n
        for z in set(zone):
            # Add new zone and update the mapping
            z_new = system.add(
                'Zone',
                param_dict=dict(idx=n_zone + 1, name=f'{n_zone + 1}')
            )
            zone_map[z] = z_new
            n_zone += 1

        # Update the zone values in the system
        system.Bus.zone.v = [zone_map[z] for z in zone]
    return ret


def write_raw(system, outfile: str, overwrite: bool = None):
    """
    Convert AMS system to PSS/E RAW file.

    Parameters
    ----------
    system : System
        The AMS system to be converted.
    outfile : str
        The output file path.
    overwrite : bool, optional
        If True, overwrite the file if it exists. If False, do not overwrite.
    """
    if not confirm_overwrite(outfile, overwrite=overwrite):
        return False

    mva = system.config.mva
    freq = system.config.freq
    name = system.files.name

    with open(outfile, 'w') as f:
        # PSS/E version and date
        f.write(f"0, {mva:.2f}, 33, 0, 1, {freq:.2f}     ")

        f.write(f"/ PSS/E 33 RAW, {report_time}\n")
        f.write(f"Created by AMS {version}\n")
        f.write(f"{copyright_msg}\n")
        f.write(f"{nowarranty_msg}\n")
        f.write(f"{name.upper()} \n")

        # --- Bus ---
        bus = system.Bus.cache.df_in
        # 0,  1,    2,      3,    4,    5,    6,     7,  8,  9,    10
        # ID, Name, BaseKV, Type, Area, Zone, Owner, Vm, Va, Vmax, Vmin
        # Define column widths for alignment
        column_widths = [6, 8, 8, 2, 3, 3, 3, 8, 8, 8, 8]
        bus_owner = {}
        for i, o in enumerate(set(bus['owner'])):
            bus_owner[o] = i + 1
        for row in bus.itertuples():
            # Prepare the data for each column
            data = [
                int(row.idx), row.name, row.Vn, int(row.type),
                int(system.Collection.idx2uid(row.area) + 1),
                int(system.Collection.idx2uid(row.zone) + 1),
                int(bus_owner[row.owner]),
                float(row.v0), float(row.a0 * rad2deg),
                float(row.vmax), float(row.vmin)]
            # Format each column with ',' as the delimiter
            formatted_row = ",".join(
                f"{value:>{width}}" if isinstance(value, (int, float)) else f"'{value:>{width - 2}}'"
                for value, width in zip(data, column_widths)
            ) + "\n"
            # Write the formatted row to the file
            f.write(formatted_row)

        # --- Load ---
        f.write("0 / END OF BUS DATA, BEGIN LOAD DATA\n")
        load = system.PQ.cache.df_in
        # 0,   1,  2,      3,    4,    5,      6,       7,  8,  9,  10, 11
        # Bus, Id, Status, Area, Zone, PL(MW), QL (MW), IP, IQ, YP, YQ, Owner
        # NOTE: load are converted to constant load, IP, IQ, YP, YQ are ignored by setting to 0
        column_widths = [6, 8, 2, 3, 3, 10, 10, 2, 2, 10, 10, 3]
        for row in load.itertuples():
            # Prepare the data for each column
            data = [
                int(row.bus),  # Bus number
                int(system.PQ.idx2uid(row.idx) + 1),  # Load ID (unique index + 1)
                int(row.u),  # Status
                int(system.Collection.idx2uid(system.PQ.get('area', row.idx, 'v')) + 1),  # Area number
                int(system.Collection.idx2uid(system.PQ.get('zone', row.idx, 'v')) + 1),  # Zone number
                float(row.p0 * mva),  # PL (MW)
                float(row.q0 * mva),  # QL (MVar)
                0.0,  # IP (ignored, set to 0)
                0.0,  # IQ (ignored, set to 0)
                0.0,  # YP (ignored, set to 0)
                0.0,  # YQ (ignored, set to 0)
                int(bus_owner[row.owner])  # Owner
            ]
            # Format each column with ',' as the delimiter
            formatted_row = ",".join(
                f"{value:>{width}}" if isinstance(value, (int, float)) else f"'{value:>{width - 2}}'"
                for value, width in zip(data, column_widths)
            ) + "\n"
            # Write the formatted row to the file
            f.write(formatted_row)

        # --- Fixed Shunt ---
        f.write("0 / END OF LOAD DATA, BEGIN FIXED SHUNT DATA\n")
        shunt = system.Shunt.cache.df_in
        # 0,   1,  2,      3,      4
        # Bus, Id, Status, G (MW), B (Mvar)
        # NOTE: ANDES parse v33 swshunt into fixed shunt
        column_widths = [6, 8, 2, 10, 10]
        for row in shunt.itertuples():
            # Prepare the data for each column
            data = [
                int(row.bus),                                   # Bus number
                int(system.Shunt.idx2uid(row.idx) + 1),         # Shunt ID (unique index + 1)
                int(row.u),                                     # Status
                float(row.g * mva),                             # Conductance (MW at system base)
                float(row.b * mva)                              # Susceptance (Mvar at system base)
            ]

            # Format each column with ',' as the delimiter
            formatted_row = ",".join(
                f"{value:>{width}}" if isinstance(value, (int, float)) else f"'{value:>{width - 2}}'"
                for value, width in zip(data, column_widths)
            ) + "\n"
            # Write the formatted row to the file
            f.write(formatted_row)

        # --- Generator ---
        f.write("0 / END OF FIXED SHUNT DATA, BEGIN GENERATOR DATA\n")
        pv = system.PV.cache.df_in
        slack = system.Slack.cache.df_in

        gen = pd.concat([pv, slack.drop(columns=['a0'])], axis=0)
        gen["subidx"] = gen.groupby('bus').cumcount() + 1

        column_widths = [6, 8, 2, 3, 3, 3, 3, 8, 8, 8, 8, 8, 8]
        # 0,  1,  2,  3,  4,  5,  6,    7,     8,  9, 10, 11, 12,   13,   14,
        # I, ID, PG, QG, QT, QB, VS, IREG, MBASE, ZR, ZX, RT, XT, GTAP, STAT,
        #    15, 16, 17, 18, 19, ...,           26,  27
        # RMPCT, PT, PB, O1, F1, ..., O4, F4, WMOD, WPF
        # The columns above for v33 is different from the manual of v34.5, which includes two new columns:
        # `NREG`` at 8 and `BSLOD` before `O1`
        for row in gen.itertuples():
            # Prepare the data for each column
            data = [
                int(row.bus),                  # I: Bus number
                int(row.subidx),               # ID: Generator ID (subindex)
                float(row.p0 * mva),           # PG: Generated MW
                float(row.q0 * mva),           # QG: Generated MVar
                float(row.qmax * mva),         # QT: Max Q (MVar)
                float(row.qmin * mva),         # QB: Min Q (MVar)
                float(row.v0),                 # VS: Setpoint voltage (p.u.)
                int(0),                        # IREG: Regulated bus (not tracked)
                float(row.Sn if hasattr(row, 'Sn') else mva),  # MBASE: Machine base MVA
                float(row.ra),                 # ZR: Armature resistance (p.u.)
                float(row.xs),                 # ZX: Synchronous reactance (p.u.)
                float(0),                      # RT: Step-up transformer resistance (not tracked)
                float(0),                      # XT: Step-up transformer reactance (not tracked)
                int(0),                        # GTAP: Step-up transformer off-nominal turns ratio (not tracked)
                int(row.u)                     # STAT: Status
            ]
            # Format each column with ',' as the delimiter
            formatted_row = ",".join(
                f"{value:>{width}}" if isinstance(value, (int, float)) else f"'{value:>{width - 2}}'"
                for value, width in zip(data, column_widths)
            ) + "\n"
            # Write the formatted row to the file
            f.write(formatted_row)

        # --- Line ---
        f.write("0 / END OF GENERATOR DATA, BEGIN BRANCH DATA\n")
        line = system.Line.cache.df_in
        branch = line[line['trans'] == 0].reset_index(drop=True)
        transf = line[line['trans'] == 1].reset_index(drop=True)
        # 1) branch
        # 0,  1,   2,   3,   4,   5,     6,     7,     8,   9,  10, 11, 12, 13,  14, 15, 16
        # I,  J,  CKT,  R,   X,   B,  RATEA, RATEB, RATEC, GI,  BI, GJ, BJ, ST, LEN, O1, F1, ..., O4, F4
        column_widths = [6, 6, 4, 10, 10, 10, 8, 8, 8, 8, 8, 8, 8, 3, 8]
        for row in branch.itertuples():
            data = [
                int(row.bus1),                      # I: From bus number
                int(row.bus2),                      # J: To bus number
                "'1'",                              # CKT: Circuit ID (default '1')
                float(row.r),                       # R: Resistance (p.u.)
                float(row.x),                       # X: Reactance (p.u.)
                float(row.b),                       # B: Total line charging (p.u.)
                float(row.rate_a),                  # RATEA: Rating A (MVA)
                float(row.rate_b),                  # RATEB: Rating B (MVA)
                float(row.rate_c),                  # RATEC: Rating C (MVA)
                0.0,                                # GI: From bus conductance (not tracked)
                0.0,                                # BI: From bus susceptance (not tracked)
                0.0,                                # GJ: To bus conductance (not tracked)
                0.0,                                # BJ: To bus susceptance (not tracked)
                int(row.u),                         # ST: Status
                0.0                                 # LEN: Line length (not tracked)
                # O1, F1, ..., O4, F4 omitted for v33
            ]
            formatted_row = ",".join(
                f"{value:>{width}}" if isinstance(value, (int, float)) else f"{value:>{width}}"
                for value, width in zip(data, column_widths)
            ) + "\n"
            f.write(formatted_row)
        # 2) transformer
        f.write("0 / END OF BRANCH DATA, BEGIN TRANSFORMER DATA\n")
        for row in transf.itertuples():
            # Map AMS Line (trans) fields to PSSE RAW transformer columns
            # Only 2-winding transformers are supported here
            # 0,  1,   2,   3,   4,   5,     6,     7,     8,   9,  10, 11, 12, 13,  14, 15, 16
            # I,  J,  CKT,  R,   X,   B,  RATEA, RATEB, RATEC, GI,  BI, GJ, BJ, ST, LEN
            data = [
                int(row.bus1),                      # I: From bus number
                int(row.bus2),                      # J: To bus number
                "'1'",                              # CKT: Circuit ID (default '1')
                float(row.r),                       # R: Resistance (p.u.)
                float(row.x),                       # X: Reactance (p.u.)
                float(row.b),                       # B: Total line charging (p.u.)
                float(row.rate_a),                  # RATEA: Rating A (MVA)
                float(row.rate_b),                  # RATEB: Rating B (MVA)
                float(row.rate_c),                  # RATEC: Rating C (MVA)
                0.0,                                # GI: From bus conductance (not tracked)
                0.0,                                # BI: From bus susceptance (not tracked)
                0.0,                                # GJ: To bus conductance (not tracked)
                0.0,                                # BJ: To bus susceptance (not tracked)
                int(row.u),                         # ST: Status
                0.0                                 # LEN: Line length (not tracked)
                # O1, F1, ..., O4, F4 omitted for v33
            ]
            formatted_row = ",".join(
                f"{value:>{width}}" if isinstance(value, (int, float)) else f"{value:>{width}}"
                for value, width in zip(data, column_widths)
            ) + "\n"
            f.write(formatted_row)

            # --- Area ---
            f.write("0 / END OF TRANSFORMER DATA, BEGIN AREA DATA\n")
            area = system.Area.cache.df_in
            for row in area.itertuples():
                # PSSE expects: ID, ISW, PDES, PTOL, NAME
                # Here, ISW, PDES, PTOL are set to 0 by default
                f.write(f"{int(system.Area.idx2uid(row.idx) + 1):6d}, 0, 0.0, 0.0, '{row.name}'\n")

            # --- Zone ---
            f.write("0 / END OF AREA DATA, BEGIN ZONE DATA\n")
            zone = system.Zone.cache.df_in
            # 0,    1
            # ID, Name
            for row in zone.itertuples():
                f.write(f"{int(system.Zone.idx2uid(row.idx) + 1):6d}, '{row.name}'\n")
            f.write("0 / END OF ZONE DATA\n")

        # End of file
        f.write("Q\n")

    return True



"""
Test system report.
"""
import unittest
import os

import ams
from ams.shared import skip_unittest_without_MISOCP


import logging

logger = logging.getLogger(__name__)


class TestReport(unittest.TestCase):
    """
    Tests for Report class.
    """

    def setUp(self) -> None:
        self.ss = ams.main.load(
            ams.get_case("5bus/pjm5bus_demo.json"),
            default_config=True,
            no_output=True,
        )
        self.expected_report = 'pjm5bus_demo_out.txt'

    def test_no_output(self):
        """
        Test no output.
        """
        self.assertTrue(self.ss.files.no_output)
        self.assertFalse(self.ss.report())

    def test_no_report(self):
        """
        Test report with no solved routine.
        """
        self.ss.files.no_output = False
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertNotIn("DCOPF", file_contents)

        os.remove(self.expected_report)

    def test_DCOPF_report(self):
        """
        Test report with DCOPF solved.
        """
        self.ss.files.no_output = False
        self.ss.DCOPF.run(solver='CLARABEL')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("DCOPF", file_contents)
        os.remove(self.expected_report)

        self.ss.DCOPF.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_DCOPF.csv'))
        os.remove('pjm5bus_demo_DCOPF.csv')

    def test_DCPF_report(self):
        """
        Test report with DCPF solved.
        """
        self.ss.files.no_output = False
        self.ss.DCPF.run(solver='CLARABEL')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("DCPF", file_contents)
        os.remove(self.expected_report)

        self.ss.DCPF.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_DCPF.csv'))
        os.remove('pjm5bus_demo_DCPF.csv')

    def test_RTED_report(self):
        """
        Test report with RTED solved.
        """
        self.ss.files.no_output = False
        self.ss.RTED.run(solver='CLARABEL')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("RTED", file_contents)
        os.remove(self.expected_report)

        self.ss.RTED.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_RTED.csv'))
        os.remove('pjm5bus_demo_RTED.csv')

    def test_RTEDDG_report(self):
        """
        Test report with RTEDDG solved.
        """
        self.ss.files.no_output = False
        self.ss.RTEDDG.run(solver='CLARABEL')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("RTEDDG", file_contents)
        os.remove(self.expected_report)

        self.ss.RTEDDG.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_RTEDDG.csv'))
        os.remove('pjm5bus_demo_RTEDDG.csv')

    @skip_unittest_without_MISOCP
    def test_RTEDES_report(self):
        """
        Test report with RTEDES solved.
        """
        self.ss.files.no_output = False
        self.ss.RTEDES.run(solver='SCIP')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("RTEDES", file_contents)
        os.remove(self.expected_report)

        self.ss.RTEDES.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_RTEDES.csv'))
        os.remove('pjm5bus_demo_RTEDES.csv')

    def test_ED_report(self):
        """
        Test report with ED solved.
        """
        self.ss.files.no_output = False
        self.ss.ED.run(solver='CLARABEL')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("ED", file_contents)
        os.remove(self.expected_report)

        self.ss.ED.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_ED.csv'))
        os.remove('pjm5bus_demo_ED.csv')

    def test_EDDG_report(self):
        """
        Test report with EDDG solved.
        """
        self.ss.files.no_output = False
        self.ss.EDDG.run(solver='CLARABEL')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("EDDG", file_contents)
        os.remove(self.expected_report)

        self.ss.EDDG.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_EDDG.csv'))
        os.remove('pjm5bus_demo_EDDG.csv')

    @skip_unittest_without_MISOCP
    def test_EDES_report(self):
        """
        Test report with EDES solved.
        """
        self.ss.files.no_output = False
        self.ss.EDES.run(solver='SCIP')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("EDES", file_contents)
        os.remove(self.expected_report)

        self.ss.EDES.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_EDES.csv'))
        os.remove('pjm5bus_demo_EDES.csv')

    @skip_unittest_without_MISOCP
    def test_UC_report(self):
        """
        Test report with UC solved.
        """
        self.ss.files.no_output = False
        self.ss.UC.run(solver='SCIP')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("UC", file_contents)
        os.remove(self.expected_report)

        self.ss.UC.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_UC.csv'))
        os.remove('pjm5bus_demo_UC.csv')

    @skip_unittest_without_MISOCP
    def test_UCDG_report(self):
        """
        Test report with UCDG solved.
        """
        self.ss.files.no_output = False
        self.ss.UCDG.run(solver='SCIP')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("UCDG", file_contents)
        os.remove(self.expected_report)

        self.ss.UCDG.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_UCDG.csv'))
        os.remove('pjm5bus_demo_UCDG.csv')

    @skip_unittest_without_MISOCP
    def test_UCES_report(self):
        """
        Test report with UCES solved.
        """
        self.ss.files.no_output = False
        self.ss.UCES.run(solver='SCIP')
        self.assertTrue(self.ss.report())
        self.assertTrue(os.path.exists(self.expected_report))

        with open(self.expected_report, "r") as report_file:
            file_contents = report_file.read()
            self.assertIn("UCES", file_contents)
        os.remove(self.expected_report)

        self.ss.UCES.export_csv()
        self.assertTrue(os.path.exists('pjm5bus_demo_UCES.csv'))
        os.remove('pjm5bus_demo_UCES.csv')



from collections import OrderedDict

from andes.core.param import NumParam, ExtParam
from andes.models.static.pq import PQData

from ams.core.model import Model


class PQ(PQData, Model):
    """
    PQ load model.

    TODO: implement type conversion in config
    """

    def __init__(self, system, config):
        PQData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'StaticLoad'
        self.config.add(OrderedDict((('pq2z', 1),
                                     ('p2p', 0.0),
                                     ('p2i', 0.0),
                                     ('p2z', 1.0),
                                     ('q2q', 0.0),
                                     ('q2i', 0.0),
                                     ('q2z', 1.0),
                                     )))
        self.config.add_extra("_help",
                              pq2z="pq2z conversion if out of voltage limits",
                              p2p="P constant power percentage for TDS. Must have (p2p+p2i+p2z)=1",
                              p2i="P constant current percentage",
                              p2z="P constant impedance percentage",
                              q2q="Q constant power percentage for TDS. Must have (q2q+q2i+q2z)=1",
                              q2i="Q constant current percentage",
                              q2z="Q constant impedance percentage",
                              )
        self.config.add_extra("_alt",
                              pq2z="(0, 1)",
                              p2p="float",
                              p2i="float",
                              p2z="float",
                              q2q="float",
                              q2i="float",
                              q2z="float",
                              )
        self.config.add_extra("_tex",
                              pq2z="z_{pq2z}",
                              p2p=r"\gamma_{p2p}",
                              p2i=r"\gamma_{p2i}",
                              p2z=r"\gamma_{p2z}",
                              q2q=r"\gamma_{q2q}",
                              q2i=r"\gamma_{q2i}",
                              q2z=r"\gamma_{q2z}",
                              )

        self.ctrl = NumParam(default=1,
                             info="load controllability",
                             tex_name=r'c_{trl}',)

        self.area = ExtParam(model='Bus', src='area', indexer=self.bus, export=False,
                             info='Retrieved area idx', vtype=str, default=None,
                             )
        self.zone = ExtParam(model='Bus', src='zone', indexer=self.bus, export=False,
                             info='Retrieved zone idx', vtype=str, default=None,
                             )



"""
Power flow routines using ANDES.
"""
import logging
from typing import Optional, Union, Type
from collections import OrderedDict

import numpy as np

from andes.utils.misc import elapsed

from ams.core.param import RParam
from ams.routines.routine import RoutineBase
from ams.opt import Var, Expression, Objective
from ams.interface import to_andes_pflow, sync_adsys

logger = logging.getLogger(__name__)


class PFlow(RoutineBase):
    """
    Power flow analysis using ANDES PFlow routine.

    More settings can be changed via ``PFlow2._adsys.config`` and ``PFlow2._adsys.PFlow.config``.

    All generator output powers, bus voltages, and angles are included in the variable definitions.
    However, not all of these are unknowns; the definitions are provided for easy access.

    References
    ----------
    1. M. L. Crow, Computational methods for electric power systems. 2015.
    2. ANDES Documentation - Simulation and Plot.
       https://docs.andes.app/en/latest/_examples/ex1.html
    """

    def __init__(self, system, config):
        RoutineBase.__init__(self, system, config)
        self.info = 'AC Power Flow'
        self.type = 'PF'
        self._adsys = None

        self.config.add(OrderedDict((('tol', 1e-6),
                                     ('max_iter', 25),
                                     ('method', 'NR'),
                                     ('check_conn', 1),
                                     ('n_factorize', 4),
                                     )))
        self.config.add_extra("_help",
                              tol="convergence tolerance",
                              max_iter="max. number of iterations",
                              method="calculation method",
                              check_conn='check connectivity before power flow',
                              n_factorize="first N iterations to factorize Jacobian in dishonest method",
                              )
        self.config.add_extra("_alt",
                              tol="float",
                              method=("NR", "dishonest", "NK"),
                              check_conn=(0, 1),
                              max_iter=">=10",
                              n_factorize=">0",
                              )

        self.Bf = RParam(info='Bf matrix',
                         name='Bf', tex_name=r'B_{f}',
                         model='mats', src='Bf',
                         no_parse=True, sparse=True,)
        self.Pfinj = RParam(info='Line power injection vector',
                            name='Pfinj', tex_name=r'P_{f}^{inj}',
                            model='mats', src='Pfinj',
                            no_parse=True,)

        self.pg = Var(info='Gen active power',
                      unit='p.u.',
                      name='pg', tex_name=r'p_g',
                      model='StaticGen', src='p')
        self.qg = Var(info='Gen reactive power',
                      unit='p.u.',
                      name='qg', tex_name=r'q_g',
                      model='StaticGen', src='q')
        self.aBus = Var(info='Bus voltage angle',
                        unit='rad',
                        name='aBus', tex_name=r'\theta_{bus}',
                        model='Bus', src='a',)
        self.vBus = Var(info='Bus voltage magnitude',
                        unit='p.u.',
                        name='vBus', tex_name=r'V_{bus}',
                        model='Bus', src='v',)
        self.plf = Expression(info='Line flow',
                              name='plf', tex_name=r'p_{lf}',
                              unit='p.u.',
                              e_str='Bf@aBus + Pfinj',
                              model='Line', src=None,)

        self.obj = Objective(name='obj',
                             info='place holder', unit='$',
                             sense='min', e_str='0',)

    def init(self, **kwargs):
        """
        Initialize the ANDES PFlow routine.

        kwargs go to andes.system.System().
        """
        self._adsys = to_andes_pflow(self.system,
                                     no_output=self.system.files.no_output,
                                     config=self.config.as_dict(),
                                     **kwargs)
        self._adsys.setup()
        self.om.init()
        self.initialized = True
        return self.initialized

    def solve(self, **kwargs):
        """
        Placeholder.
        """
        return True

    def run(self, **kwargs):
        """
        Run the routine.
        """
        if not self.initialized:
            self.init()

        t0, _ = elapsed()
        _ = self._adsys.PFlow.run()
        self.exit_code = self._adsys.exit_code
        self.converged = self.exit_code == 0
        _, s = elapsed(t0)
        self.exec_time = float(s.split(" ")[0])

        self.unpack(res=None)
        return self.converged

    def _post_solve(self):
        """
        Placeholder.
        """
        return True

    def unpack(self, res, **kwargs):
        """
        Unpack the results from ANDES PFlow routine.
        """
        # TODO: maybe also include the DC devices results
        sys = self.system
        # --- device results ---
        bus_idx = sys.Bus.idx.v
        sys.Bus.set(src='v', attr='v', idx=bus_idx,
                    value=self._adsys.Bus.get(src='v', attr='v', idx=bus_idx))
        sys.Bus.set(src='a', attr='v', idx=bus_idx,
                    value=self._adsys.Bus.get(src='a', attr='v', idx=bus_idx))
        pv_idx = sys.PV.idx.v
        pv_u = sys.PV.get(src='u', attr='v', idx=pv_idx)
        # NOTE: for p, we should consider the online status as p0 is a param
        sys.PV.set(src='p', attr='v', idx=pv_idx,
                   value=pv_u * sys.PV.get(src='p0', attr='v', idx=pv_idx))
        sys.PV.set(src='q', attr='v', idx=pv_idx,
                   value=self._adsys.PV.get(src='q', attr='v', idx=pv_idx))
        slack_idx = sys.Slack.idx.v
        sys.Slack.set(src='p', attr='v', idx=slack_idx,
                      value=self._adsys.Slack.get(src='p', attr='v', idx=slack_idx))
        sys.Slack.set(src='q', attr='v', idx=slack_idx,
                      value=self._adsys.Slack.get(src='q', attr='v', idx=slack_idx))
        # --- routine results ---
        self.pg.optz.value = sys.StaticGen.get(src='p', attr='v', idx=self.pg.get_all_idxes())
        self.qg.optz.value = sys.StaticGen.get(src='q', attr='v', idx=self.qg.get_all_idxes())
        self.aBus.optz.value = sys.Bus.get(src='a', attr='v', idx=self.aBus.get_all_idxes())
        self.vBus.optz.value = sys.Bus.get(src='v', attr='v', idx=self.vBus.get_all_idxes())
        return True

    def update(self, params=None, build_mats=False):
        """
        This method updates the parameters in the optimization model. In this routine,
        the `params` and `build_mats` arguments are not used because the parameters
        are updated directly to the ANDES system.
        """
        if not self.initialized:
            self.init()

        sync_adsys(self.system, self._adsys)

        return True

    def enable(self, name):
        raise NotImplementedError

    def disable(self, name):
        raise NotImplementedError

    def addRParam(self,
                  name: str,
                  tex_name: Optional[str] = None,
                  info: Optional[str] = None,
                  src: Optional[str] = None,
                  unit: Optional[str] = None,
                  model: Optional[str] = None,
                  v: Optional[np.ndarray] = None,
                  indexer: Optional[str] = None,
                  imodel: Optional[str] = None,):
        """
        Not supported!
        """
        raise NotImplementedError

    def addService(self,
                   name: str,
                   value: np.ndarray,
                   tex_name: str = None,
                   unit: str = None,
                   info: str = None,
                   vtype: Type = None,):
        """
        Not supported!
        """
        raise NotImplementedError

    def addConstrs(self,
                   name: str,
                   e_str: str,
                   info: Optional[str] = None,
                   is_eq: Optional[str] = False,):
        """
        Not supported!
        """
        raise NotImplementedError

    def addVars(self,
                name: str,
                model: Optional[str] = None,
                shape: Optional[Union[int, tuple]] = None,
                tex_name: Optional[str] = None,
                info: Optional[str] = None,
                src: Optional[str] = None,
                unit: Optional[str] = None,
                horizon: Optional[RParam] = None,
                nonneg: Optional[bool] = False,
                nonpos: Optional[bool] = False,
                cplx: Optional[bool] = False,
                imag: Optional[bool] = False,
                symmetric: Optional[bool] = False,
                diag: Optional[bool] = False,
                psd: Optional[bool] = False,
                nsd: Optional[bool] = False,
                hermitian: Optional[bool] = False,
                boolean: Optional[bool] = False,
                integer: Optional[bool] = False,
                pos: Optional[bool] = False,
                neg: Optional[bool] = False,):
        """
        Not supported!
        """
        raise NotImplementedError



"""
EV Aggregator module.

EVD is the generated datasets, and EVA is the aggregator model.

Reference:
[1] J. Wang et al., "Electric Vehicles Charging Time Constrained Deliverable Provision of Secondary
Frequency Regulation," in IEEE Transactions on Smart Grid, doi: 10.1109/TSG.2024.3356948.
[2] M. Wang, Y. Mu, Q. Shi, H. Jia and F. Li, "Electric Vehicle Aggregator Modeling and Control for
Frequency Regulation Considering Progressive State Recovery," in IEEE Transactions on Smart Grid,
vol. 11, no. 5, pp. 4176-4189, Sept. 2020, doi: 10.1109/TSG.2020.2981843.
"""

import logging
import itertools
from collections import OrderedDict

import scipy.stats as stats

from andes.core.param import NumParam
from andes.core.model import ModelData
from andes.shared import np, pd
from andes.utils.misc import elapsed

from ams.core.model import Model
from ams.utils.paths import ams_root
from ams.core import Config

logger = logging.getLogger(__name__)


# NOTE: following definition comes from ref[2], except `tt` that is assumed by ref[1]
# normal distribution parameters
ndist = {'soci': {'mu': 0.3, 'var': 0.05, 'lb': 0.2, 'ub': 0.4},
         'socd': {'mu': 0.8, 'var': 0.03, 'lb': 0.7, 'ub': 0.9},
         'ts1': {'mu': -6.5, 'var': 3.4, 'lb': 0.0, 'ub': 5.5},
         'ts2': {'mu': 17.5, 'var': 3.4, 'lb': 5.5, 'ub': 24.0},
         'tf1': {'mu': 8.9, 'var': 3.4, 'lb': 0.0, 'ub': 20.9},
         'tf2': {'mu': 32.9, 'var': 3.4, 'lb': 20.9, 'ub': 24.0},
         'tt': {'mu': 0.5, 'var': 0.02, 'lb': 0, 'ub': 1}}
# uniform distribution parameters
udist = {'Pc': {'lb': 5.0, 'ub': 7.0},
         'Pd': {'lb': 5.0, 'ub': 7.0},
         'nc': {'lb': 0.88, 'ub': 0.95},
         'nd': {'lb': 0.88, 'ub': 0.95},
         'Q': {'lb': 20.0, 'ub': 30.0}}


class EVD(ModelData, Model):
    """
    In the EVD, each single EV is recorded as a device with its own parameters.
    The parameters are generated from given statistical distributions.
    """

    def __init__(self, N=10000, Ns=20, Tagc=4, SOCf=0.2, r=0.5,
                 t=18, seed=None, name='EVA', A_csv=None):
        """
        Initialize the EV aggregation model.

        Parameters
        ----------
        N : int, optional
            Number of related EVs, default is 10000.
        Ns : int, optional
            Number of SOC intervals, default is 20.
        Tagc : int, optional
            AGC time intervals in seconds, default is 4.
        SOCf : float, optional
            Force charge SOC level between 0 and 1, default is 0.2.
        r : float, optional
            Ratio of time range 1 to time range 2 between 0 and 1, default is 0.5.
        seed : int or None, optional
            Seed for random number generator, default is None.
        t : int, optional
            Current time in 24 hours, default is 18.
        name : str, optional
            Name of the EVA, default is 'EVA'.
        A_csv : str, optional
            Path to the CSV file containing the state space matrix A, default is None.
        """
        # inherit attributes and methods from ANDES `ModelData` and AMS `Model`
        ModelData.__init__(self)
        Model.__init__(self, system=None, config=None)

        self.evdname = name

        # internal flags
        self.is_setup = False        # if EVA has been setup

        self.t = np.array(t, dtype=float)  # time in 24 hours
        self.eva = None  # EV Aggregator
        self.A_csv = A_csv  # path to the A matrix

        # manually set config as EVA is not processed by the system
        self.config = Config(self.__class__.__name__)
        self.config.add(OrderedDict((('n', int(N)),
                                     ('ns', Ns),
                                     ('tagc', Tagc),
                                     ('socf', SOCf),
                                     ('r', r),
                                     ('socl', 0),
                                     ('socu', 1),
                                     ('tf', self.t),
                                     ('prumax', 0),
                                     ('prdmax', 0),
                                     ('seed', seed),
                                     )))
        self.config.add_extra("_help",
                              n="Number of related EVs",
                              ns="SOC intervals",
                              tagc="AGC time intervals in seconds",
                              socf="Force charge SOC level",
                              r="ratio of time range 1 to time range 2",
                              socl="lowest SOC limit",
                              socu="highest SOC limit",
                              tf="EVA running end time in 24 hours",
                              prumax="maximum power of regulation up, in MW",
                              prdmax="maximum power of regulation down, in MW",
                              seed='seed (or None) for random number generator',
                              )
        self.config.add_extra("_tex",
                              n='N_{ev}',
                              ns='N_s',
                              tagc='T_{agc}',
                              socf='SOC_f',
                              r='r',
                              socl='SOC_{l}',
                              socu='SOC_{u}',
                              tf='T_f',
                              prumax='P_{ru,max}',
                              prdmax='P_{rd,max}',
                              seed='seed',
                              )
        self.config.add_extra("_alt",
                              n='int',
                              ns="int",
                              tagc="float",
                              socf="float",
                              r="float",
                              socl="float",
                              socu="float",
                              tf="float",
                              prumax="float",
                              prdmax="float",
                              seed='int or None',
                              )

        unit = self.config.socu / self.config.ns
        self.soc_intv = OrderedDict({
            i: (np.around(i * unit, 2), np.around((i + 1) * unit, 2))
            for i in range(self.config.ns)
        })

        # NOTE: the parameters and variables are declared here and populated in `setup()`
        # param `idx`, `name`, and `u` are already included in `ModelData`
        # variables here are actually declared as parameters for memory saving
        # because ams.core.var.Var has more overhead

        # --- parameters ---
        self.namax = NumParam(default=0,
                              info='maximum number of action')
        self.ts = NumParam(default=0, vrange=(0, 24),
                           info='arrive time, in 24 hours')
        self.tf = NumParam(default=0, vrange=(0, 24),
                           info='departure time, in 24 hours')
        self.tt = NumParam(default=0,
                           info='Tolerance of increased charging time, in hours')
        self.soci = NumParam(default=0,
                             info='initial SOC')
        self.socd = NumParam(default=0,
                             info='demand SOC')
        self.Pc = NumParam(default=0,
                           info='rated charging power, in kW')
        self.Pd = NumParam(default=0,
                           info='rated discharging power, in kW')
        self.nc = NumParam(default=0,
                           info='charging efficiency',
                           vrange=(0, 1))
        self.nd = NumParam(default=0,
                           info='discharging efficiency',
                           vrange=(0, 1))
        self.Q = NumParam(default=0,
                          info='rated capacity, in kWh')

        # --- variables ---
        self.soc0 = NumParam(default=0,
                             info='previous SOC')
        self.u0 = NumParam(default=0,
                           info='previous online status')
        self.na0 = NumParam(default=0,
                            info='previous action number')
        self.soc = NumParam(default=0,
                            info='SOC')
        self.na = NumParam(default=0,
                           info='action number')

    def setup(self, ndist=ndist, udist=udist):
        """
        Setup the EV aggregation model.

        Parameters
        ----------
        ndist : dict, optional
            Normal distribution parameters, default by built-in `ndist`.
        udist : dict, optional
            Uniform distribution parameters, default by built-in `udist`.

        Returns
        -------
        is_setup : bool
            If the setup is successful.
        """
        if self.is_setup:
            logger.warning(f'{self.evdname} aggregator has been setup, setup twice is not allowed.')
            return False

        t0, _ = elapsed()

        # manually set attributes as EVA is not processed by the system
        self.n = self.config.n
        self.idx.v = ['SEV_' + str(i+1) for i in range(self.config.n)]
        self.name.v = ['SEV ' + str(i+1) for i in range(self.config.n)]
        self.u.v = np.array(self.u.v, dtype=int)
        self.uid = {self.idx.v[i]: i for i in range(self.config.n)}

        # --- populate parameters' value ---
        # set `soci`, `socd`, `tt`
        self.soci.v = build_truncnorm(ndist['soci']['mu'], ndist['soci']['var'],
                                      ndist['soci']['lb'], ndist['soci']['ub'],
                                      self.config.n, self.config.seed)
        self.socd.v = build_truncnorm(ndist['socd']['mu'], ndist['socd']['var'],
                                      ndist['socd']['lb'], ndist['socd']['ub'],
                                      self.config.n, self.config.seed)
        self.tt.v = build_truncnorm(ndist['tt']['mu'], ndist['tt']['var'],
                                    ndist['tt']['lb'], ndist['tt']['ub'],
                                    self.config.n, self.config.seed)
        # set `ts`, `tf`
        tdf = pd.DataFrame({
            col: build_truncnorm(ndist[col]['mu'], ndist[col]['var'],
                                 ndist[col]['lb'], ndist[col]['ub'],
                                 self.config.n, self.config.seed)
            for col in ['ts1', 'ts2', 'tf1', 'tf2']
        })

        nev_t1 = int(self.config.n * self.config.r)  # number of EVs in time range 1
        tp1 = tdf[['ts1', 'tf1']].sample(n=nev_t1, random_state=self.config.seed)
        tp2 = tdf[['ts2', 'tf2']].sample(n=self.config.n-nev_t1, random_state=self.config.seed)
        tp = pd.concat([tp1, tp2], axis=0).reset_index(drop=True).fillna(0)
        tp['ts'] = tp['ts1'] + tp['ts2']
        tp['tf'] = tp['tf1'] + tp['tf2']
        # Swap ts and tf if ts > tf
        check = tp['ts'] > tp['tf']
        tp.loc[check, ['ts', 'tf']] = tp.loc[check, ['tf', 'ts']].values

        self.ts.v = tp['ts'].values
        self.tf.v = tp['tf'].values

        # set `Pc`, `Pd`, `nc`, `nd`, `Q`
        # NOTE: here it assumes (1) Pc == Pd, (2) nc == nd given by ref[2]
        if self.config.seed is not None:
            np.random.seed(self.config.seed)
        self.Pc.v = np.random.uniform(udist['Pc']['lb'], udist['Pc']['ub'], self.config.n)
        self.Pd.v = self.Pc.v
        self.nc.v = np.random.uniform(udist['nc']['lb'], udist['nc']['ub'], self.config.n)
        self.nd.v = self.nc.v
        self.Q.v = np.random.uniform(udist['Q']['lb'], udist['Q']['ub'], self.config.n)

        # --- adjust variables given current time ---
        self.g_u()  # update online status
        # adjust SOC considering random behavior
        # NOTE: here we ignore the AGC participation before the current time `self.t`

        # stayed time for the EVs arrived before t, reset negative time to 0
        tc = np.maximum(self.t - self.ts.v, 0)
        self.soc.v = self.soci.v + tc * self.Pc.v * self.nc.v / self.Q.v  # charge them

        tr = (self.socd.v - self.soci.v) * self.Q.v / self.Pc.v / self.nc.v  # time needed to charge to socd

        # ratio of stay/required time, stay less than required time reset to 1
        kt = np.maximum(tc / tr, 1)
        socp = self.socd.v + np.log(kt) * (1 - self.socd.v)  # log scale higher than socd
        mask = kt > 1
        self.soc.v[mask] = socp[mask]  # Update soc

        # clip soc to min/max
        self.soc.v = np.clip(self.soc.v, self.config.socl, self.config.socu)

        self.soc0.v = self.soc.v.copy()
        self.u0.v = self.u.v.copy()

        self.evd = EVA(evd=self, A_csv=self.A_csv)

        self.is_setup = True

        _, s = elapsed(t0)
        msg = f'{self.evdname} aggregator setup in {s}, and the current time is {self.t} H.\n'
        msg += f'It has {self.config.n} EVs in total and {self.u.v.sum()} EVs online.'
        logger.info(msg)

        return self.is_setup

    def g_u(self):
        """
        Update online status of EVs based on current time.
        """
        self.u.v = ((self.ts.v <= self.t) & (self.t <= self.tf.v)).astype(int)

        return True


class EVA:
    """
    State space modeling based EV aggregation model.
    """

    def __init__(self, evd, A_csv=None):
        """
        Parameters
        ----------
        EVD : ams.extension.eva.EVD
            EV Aggregator model.
        A_csv : str, optional
            Path to the CSV file containing the state space matrix A, default is None.
        """
        self.parent = evd

        # states of EV, intersection of charging status and SOC intervals
        # C: charging, I: idle, D: discharging
        states = list(itertools.product(['C', 'I', 'D'], self.parent.soc_intv.keys()))
        self.state = OrderedDict(((''.join(str(i) for i in s), 0.0) for s in states))

        # NOTE: 3*ns comes from the intersection of charging status and SOC intervals
        ns = self.parent.config.ns
        # NOTE: x, A will be updated in `setup()`
        self.x = np.zeros(3*ns)

        # A matrix
        default_A_csv = ams_root() + '/extension/Aest.csv'
        if A_csv:
            try:
                self.A = pd.read_csv(A_csv).values
                logger.debug(f'Loaded A matrix from {A_csv}.')
            except FileNotFoundError:
                self.A = pd.read_csv(default_A_csv).values
                logger.debug(f'File {A_csv} not found, using default A matrix.')
        else:
            self.A = pd.read_csv(default_A_csv).values
            logger.debug('No A matrix provided, using default A matrix.')

        mate = np.eye(ns)
        mat0 = np.zeros((ns, ns))
        self.B = np.vstack((-mate, mate, mat0))
        self.C = np.vstack((mat0, -mate, mate))

        # SSM variables
        kde = stats.gaussian_kde(self.parent.Pc.v)
        step = 0.01
        Pl_values = np.arange(self.parent.Pc.v.min(), self.parent.Pc.v.max(), step)
        self.Pave = 1e-3 * np.sum([Pl * kde.integrate_box(Pl, Pl + step) for Pl in Pl_values])  # kw to MW

        # NOTE: D, Da, Db, Dc, Dd will be scaled by Pave later in `setup()`
        vec1 = np.ones((1, ns))
        vec0 = np.zeros((1, ns))
        self.D = self.Pave * np.hstack((-vec1, vec0, vec0))
        self.Da = self.Pave * np.hstack((vec0, vec0, vec1))
        self.Db = self.Pave * np.hstack((vec1, vec1, vec1))
        self.Db[0, ns] = 0  # low charged EVs don't DC
        self.Dc = self.Pave * np.hstack((-vec1, vec0, vec0))
        self.Dd = self.Pave * np.hstack((-vec1, -vec1, -vec1))
        self.Dd[0, 2*ns-1] = 0  # overcharged EVs don't C


def build_truncnorm(mu, var, lb, ub, n, seed):
    """
    Helper function to generate truncated normal distribution
    using scipy.stats.

    Parameters
    ----------
    mu : float
        Mean of the normal distribution.
    var : float
        Variance of the normal distribution.
    lb : float
        Lower bound of the truncated distribution.
    ub : float
        Upper bound of the truncated distribution.
    n : int
        Number of samples to generate.
    seed : int
        Random seed to use.

    Returns
    -------
    samples : ndarray
        Generated samples.
    """
    a = (lb - mu) / var
    b = (ub - mu) / var
    distribution = stats.truncnorm(a, b, loc=mu, scale=var)
    return distribution.rvs(n, random_state=seed)



"""
Base class for variables.
"""

from typing import Optional
import logging

import numpy as np

logger = logging.getLogger(__name__)


class Algeb:
    """
    Algebraic variable class.

    This class is simplified from ``andes.core.var.Algeb``.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 ):
        self.name = name
        self.info = info
        self.unit = unit

        self.tex_name = tex_name if tex_name else name
        self.owner = None  # instance of the owner Model
        self.id = None     # variable internal index inside a model (assigned in run time)

        # TODO: set a
        # address into the variable and equation arrays (dae.f/dae.g and dae.x/dae.y)
        self.a: np.ndarray = np.array([], dtype=int)

        self.v: np.ndarray = np.array([], dtype=float)  # variable value array

    def __repr__(self):
        if self.owner.n == 0:
            span = []

        elif 1 <= self.owner.n <= 20:
            span = f'a={self.a}, v={self.v}'

        else:
            span = []
            span.append(self.a[0])
            span.append(self.a[-1])
            span.append(self.a[1] - self.a[0])
            span = ':'.join([str(i) for i in span])
            span = 'a=[' + span + ']'

        return f'{self.__class__.__name__}: {self.owner.__class__.__name__}.{self.name}, {span}'

    @property
    def class_name(self):
        return self.__class__.__name__



"""
Module for optimization ExpressionCalc.
"""
import logging

from typing import Optional
import re

import numpy as np

from ams.shared import sps  # NOQA

import cvxpy as cp

from ams.utils import pretty_long_message
from ams.shared import _prefix, _max_length

from ams.opt import OptzBase, ensure_symbols, ensure_mats_and_parsed

logger = logging.getLogger(__name__)


class ExpressionCalc(OptzBase):
    """
    Class for calculating expressions.

    Note that `ExpressionCalc` is not a CVXPY expression, and it should not be used
    in the optimization model.
    It is used to calculate expression values **after** successful optimization.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 e_str: Optional[str] = None,
                 model: Optional[str] = None,
                 src: Optional[str] = None,
                 ):
        OptzBase.__init__(self, name=name, info=info, unit=unit, model=model)
        self.optz = None
        self.e_str = e_str
        self.code = None
        self.src = src

    @ensure_symbols
    def parse(self):
        """
        Parse the Expression.
        """
        # parse the expression str
        sub_map = self.om.rtn.syms.sub_map
        code_expr = self.e_str
        for pattern, replacement in sub_map.items():
            try:
                code_expr = re.sub(pattern, replacement, code_expr)
            except Exception as e:
                raise Exception(f"Error in parsing expr <{self.name}>.\n{e}")
        # store the parsed expression str code
        self.code = code_expr
        msg = f" - ExpressionCalc <{self.name}>: {self.e_str}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        return True

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the expression.
        """
        msg = f" - Expression <{self.name}>: {self.code}"
        logger.debug(pretty_long_message(msg, _prefix, max_length=_max_length))
        try:
            local_vars = {'self': self, 'np': np, 'cp': cp, 'sps': sps}
            self.optz = self._evaluate_expression(self.code, local_vars=local_vars)
        except Exception as e:
            raise Exception(f"Error in evaluating ExpressionCalc <{self.name}>.\n{e}")
        return True

    def _evaluate_expression(self, code, local_vars=None):
        """
        Helper method to evaluate the expression code.

        Parameters
        ----------
        code : str
            The code string representing the expression.

        Returns
        -------
        cp.Expression
            The evaluated cvxpy expression.
        """
        return eval(code, {}, local_vars)

    @property
    def v(self):
        """
        Return the CVXPY expression value.
        """
        if self.optz is None:
            return None
        else:
            return self.optz.value

    @property
    def e(self):
        """
        Return the calculated expression value.
        """
        if self.code is None:
            logger.info(f"ExpressionCalc <{self.name}> is not parsed yet.")
            return None

        val_map = self.om.rtn.syms.val_map
        code = self.code
        for pattern, replacement in val_map.items():
            try:
                code = re.sub(pattern, replacement, code)
            except TypeError as e:
                raise TypeError(e)

        try:
            logger.debug(pretty_long_message(f"Value code: {code}",
                                             _prefix, max_length=_max_length))
            local_vars = {'self': self, 'np': np, 'cp': cp, 'val_map': val_map}
            return self._evaluate_expression(code, local_vars)
        except Exception as e:
            logger.error(f"Error in calculating expr <{self.name}>.\n{e}")
            return None



"""
Cost model.
"""

from andes.core import (ModelData, IdxParam, NumParam)

from ams.core.model import Model


class GCostData(ModelData):
    def __init__(self):
        super().__init__()
        self.gen = IdxParam(info="static generator index",
                            model='StaticGen',
                            mandatory=True,
                            )
        self.type = NumParam(default=2,
                             info='Cost model type. 1 for piecewise linear, 2 for polynomial',
                             power=False,
                             tex_name=r't_{ype}',
                             vrange=(1, 2),
                             )
        self.csu = NumParam(default=0,
                            info='startup cost in US dollars',
                            power=False,
                            tex_name=r'c_{su}',
                            unit='$',
                            )
        self.csd = NumParam(default=0,
                            info='shutdown cost in US dollars',
                            power=False,
                            tex_name=r'c_{sd}',
                            unit='$',
                            )
        self.c2 = NumParam(default=0,
                           info='coefficient 2',
                           power=False,
                           tex_name=r'c_{2}',
                           unit=r'$/(p.u.*h)^2',
                           )
        self.c1 = NumParam(default=0,
                           info='coefficient 1',
                           power=False,
                           tex_name=r'c_{1}',
                           unit=r'$/p.u.*h',
                           )
        self.c0 = NumParam(default=0,
                           info='coefficient 0',
                           power=False,
                           tex_name=r'c_{0}',
                           unit=r'$',
                           )


# TODO: double check the picewise linear part documentation
class GCost(GCostData, Model):
    """
    Generator cost model, similar to MATPOWER ``gencost`` format.

    ``type`` is the cost model type. 1 for piecewise linear, 2 for polynomial.

    In piecewise linear cost model, cost function f(p) is defined by a set of points:
    (p0, c0), (p1, c1), (p2, c2), where p0 < p1 < p2.

    In quadratic cost model, cost function f(p) is defined by a set of coefficients:
    f(p) = c2 * p^2 + c1 * p + c0.
    """

    def __init__(self, system, config):
        GCostData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Cost'


class SFRCost(ModelData, Model):
    """
    Linear SFR cost model.
    """

    def __init__(self, system, config):
        ModelData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Cost'
        self.gen = IdxParam(info="static generator index",
                            model='StaticGen',
                            mandatory=True,)
        self.cru = NumParam(default=0,
                            tex_name=r'c_{r}', unit=r'$/(p.u.*h)',
                            info='cost for RegUp reserve',)
        self.crd = NumParam(default=0,
                            tex_name=r'c_{r}', unit=r'$/(p.u.*h)',
                            info='cost for RegDn reserve',)


class SRCost(ModelData, Model):
    """
    Linear spinning reserve cost model.
    """

    def __init__(self, system, config):
        ModelData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Cost'
        self.gen = IdxParam(info="static generator index",
                            model='StaticGen',
                            mandatory=True,)
        self.csr = NumParam(default=0,
                            tex_name=r'c_{sr}', unit=r'$/(p.u.*h)',
                            info='cost for spinning reserve',)


class NSRCost(ModelData, Model):
    """
    Linear non-spinning reserve cost model.
    """

    def __init__(self, system, config):
        ModelData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Cost'
        self.gen = IdxParam(info="static generator index",
                            model='StaticGen',
                            mandatory=True,)
        self.cnsr = NumParam(default=0,
                             tex_name=r'c_{nsr}', unit=r'$/(p.u.*h)',
                             info='cost for non-spinning reserve',)


class DCost(ModelData, Model):
    """
    Linear cost model for dispatchable loads.
    """
    def __init__(self, system, config):
        ModelData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Cost'
        self.pq = IdxParam(info="static load index",
                           mandatory=True,)
        self.cdp = NumParam(default=999,
                            tex_name=r'c_{d,p}', unit=r'$/(p.u.*h)',
                            info='cost for unserve load penalty',)


class VSGCostData(ModelData):
    def __init__(self):
        super().__init__()
        self.reg = IdxParam(info="Renewable generator idx",
                            model='RenGen',
                            mandatory=True,
                            )
        self.cm = NumParam(default=0,
                           info='cost for emulated inertia (M)',
                           tex_name=r'c_{r}',
                           unit=r'$/s',
                           )
        self.cd = NumParam(default=0,
                           info='cost for emulated damping (D)',
                           tex_name=r'c_{r}',
                           unit=r'$/p.u.',
                           )


class VSGCost(VSGCostData, Model):
    """
    Linear cost model for VSG emulated inertia (M) and damping (D).
    """

    def __init__(self, system, config):
        VSGCostData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Cost'



"""
ACOPF routines.
"""

from ams.routines.pypower import ACOPF1


class ACOPF(ACOPF1):
    """
    Alias for ACOPF1.
    """

    def __init__(self, system, config):
        super().__init__(system, config)



"""
Module for optimization base classes.
"""
import logging

from typing import Optional

from ams.utils.misc import deprec_get_idx


logger = logging.getLogger(__name__)


def ensure_symbols(func):
    """
    Decorator to ensure that symbols are generated before parsing.
    If not, it runs self.rtn.syms.generate_symbols().

    Designed to be used on the `parse` method of the optimization elements (`OptzBase`)
    and optimization model (`OModel`), i.e., `Var`, `Param`, `Constraint`, `Objective`,
    and `ExpressionCalc`.

    Parsing before symbol generation can give wrong results. Ensure that symbols
    are generated before calling the `parse` method.
    """

    def wrapper(self, *args, **kwargs):
        if not self.rtn._syms:
            logger.debug(f"<{self.rtn.class_name}> symbols are not generated yet. Generating now...")
            self.rtn.syms.generate_symbols()
        return func(self, *args, **kwargs)
    return wrapper


def ensure_mats_and_parsed(func):
    """
    Decorator to ensure that system matrices are built and the OModel is parsed
    before evaluation. If not, it runs the necessary methods to initialize them.

    Designed to be used on the `evaluate` method of optimization elements (`OptzBase`)
    and the optimization model (`OModel`), i.e., `Var`, `Param`, `Constraint`, `Objective`,
    and `ExpressionCalc`.

    Evaluation before building matrices and parsing the OModel can lead to errors. Ensure that
    system matrices are built and the OModel is parsed before calling the `evaluate` method.
    """

    def wrapper(self, *args, **kwargs):
        try:
            if not self.rtn.system.mats.initialized:
                logger.debug("System matrices are not built yet. Building now...")
                self.rtn.system.mats.build()
            if isinstance(self, (OptzBase)):
                if not self.om.parsed:
                    logger.debug("OModel is not parsed yet. Parsing now...")
                    self.om.parse()
            else:
                if not self.parsed:
                    logger.debug("OModel is not parsed yet. Parsing now...")
                    self.parse()
        except Exception as e:
            logger.error(f"Error during initialization or parsing: {e}")
            raise e
        return func(self, *args, **kwargs)
    return wrapper


class OptzBase:
    """
    Base class for optimization elements.
    Ensure that symbols are generated before calling the `parse` method. Parsing
    before symbol generation can lead to incorrect results.

    Parameters
    ----------
    name : str, optional
        Name of the optimization element.
    info : str, optional
        Descriptive information about the optimization element.
    unit : str, optional
        Unit of measurement for the optimization element.

    Attributes
    ----------
    rtn : ams.routines.Routine
        The owner routine instance.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 model: Optional[str] = None,
                 ):
        self.om = None
        self.name = name
        self.info = info
        self.unit = unit
        self.is_disabled = False
        self.rtn = None
        self.optz = None  # corresponding optimization element
        self.code = None
        self.model = model  # indicate if this element belongs to a model or group
        self.owner = None  # instance of the owner model or group
        self.is_group = False

    @ensure_symbols
    def parse(self):
        """
        Parse the object.
        """
        raise NotImplementedError

    @ensure_mats_and_parsed
    def evaluate(self):
        """
        Evaluate the object.
        """
        raise NotImplementedError

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    @property
    def n(self):
        """
        Return the number of elements.
        """
        if self.owner is None:
            return len(self.v)
        else:
            return self.owner.n

    @property
    def shape(self):
        """
        Return the shape.
        """
        try:
            return self.om.__dict__[self.name].shape
        except KeyError:
            logger.warning('Shape info is not ready before initialization.')
            return None

    @property
    def size(self):
        """
        Return the size.
        """
        if self.rtn.initialized:
            return self.om.__dict__[self.name].size
        else:
            logger.warning(f'Routine <{self.rtn.class_name}> is not initialized yet.')
            return None

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.name}'

    @deprec_get_idx
    def get_idx(self):
        if self.is_group:
            return self.owner.get_all_idxes()
        elif self.owner is None:
            logger.info(f'{self.class_name} <{self.name}> has no owner.')
            return None
        else:
            return self.owner.idx.v

    def get_all_idxes(self):
        """
        Return all the indexes of this item.

        .. note::
            New in version 1.0.0.

        Returns
        -------
        list
            A list of indexes.
        """

        if self.is_group:
            return self.owner.get_all_idxes()
        elif self.owner is None:
            logger.info(f'{self.class_name} <{self.name}> has no owner.')
            return None
        else:
            return self.owner.idx.v



import logging

import numpy as np

from andes.core.param import NumParam
from andes.models.bus import BusData

from ams.core.var import Algeb
from ams.core.model import Model

logger = logging.getLogger(__name__)


class Bus(BusData, Model):
    """
    AC Bus model.
    """

    def __init__(self, system, config):
        BusData.__init__(self)
        Model.__init__(self, system, config)

        self.group = 'ACTopology'
        # NOTE: in ANDES, self.zone is defined to trace a non-existing model "Zone"
        # in AMS, model "Zone" is developed,
        # so we need to change the model name of IdxParam self.zone
        self.zone.model = 'Zone'

        self.type = NumParam(name='type',
                             info='bus type, 1=PQ, 2=PV, 3=ref, 4=isolated (place holder)',
                             default=1,
                             vtype=int,
                             )

        self.a = Algeb(name='a',
                       tex_name=r'\theta',
                       info='voltage angle',
                       unit='rad',
                       )
        self.v = Algeb(name='v',
                       tex_name='V',
                       info='voltage magnitude',
                       unit='p.u.',
                       )

        # island information
        self.n_islanded_buses = 0
        self.island_sets = list()
        self.islanded_buses = list()  # list of lists containing bus uid of islands
        self.islands = list()         # same as the above
        self.islanded_a = np.array([])
        self.islanded_v = np.array([])



"""
Unit commitment routines.
"""
import logging
from collections import OrderedDict
import numpy as np
import pandas as pd

from ams.core.param import RParam
from ams.core.service import (NumOp, NumOpDual, MinDur)
from ams.routines.dcopf import DCOPF
from ams.routines.rted import RTEDBase
from ams.routines.ed import SRBase, MPBase, ESD1MPBase, DGBase

from ams.opt import Var, Constraint

logger = logging.getLogger(__name__)


class NSRBase:
    """
    Base class for non-spinning reserve.
    """

    def __init__(self) -> None:
        self.cnsr = RParam(info='cost for non-spinning reserve',
                           name='cnsr', tex_name=r'c_{nsr}',
                           model='NSRCost', src='cnsr',
                           unit=r'$/(p.u.*h)',
                           indexer='gen', imodel='StaticGen',)
        self.dnsrp = RParam(info='non-spinning reserve requirement in percentage',
                            name='dnsr', tex_name=r'd_{nsr}',
                            model='NSR', src='demand',
                            unit='%',)
        self.prns = Var(info='non-spinning reserve',
                        name='prns', tex_name=r'p_{r, ns}',
                        model='StaticGen', nonneg=True,)

        self.dnsrpz = NumOpDual(u=self.pdz, u2=self.dnsrp, fun=np.multiply,
                                name='dnsrpz', tex_name=r'd_{nsr, p, z}',
                                info='zonal non-spinning reserve requirement in percentage',)
        self.dnsr = NumOpDual(u=self.dnsrpz, u2=self.sd, fun=np.multiply,
                              rfun=np.transpose,
                              name='dnsr', tex_name=r'd_{nsr}',
                              info='zonal non-spinning reserve requirement',
                              no_parse=True,)

        # NOTE: define e_str in dispatch model
        self.prnsb = Constraint(info='non-spinning reserve balance',
                                name='prnsb', is_eq=True,)
        self.rnsr = Constraint(info='non-spinning reserve requirement',
                               name='rnsr', is_eq=False,)


class UC(DCOPF, RTEDBase, MPBase, SRBase, NSRBase):
    """
    DC-based unit commitment (UC):
    The bilinear term in the formulation is linearized with big-M method.

    Non-negative var `pdu` is introduced as unserved load with its penalty `cdp`.

    Constraints include power balance, ramping, spinning reserve, non-spinning reserve,
    minimum ON/OFF duration.
    The cost inludes generation cost, startup cost, shutdown cost, spinning reserve cost,
    non-spinning reserve cost, and unserved load penalty.

    Method ``_initial_guess`` is used to make initial guess for commitment decision if all
    generators are online at initial. It is a simple heuristic method, which may not be optimal.

    Notes
    -----
    1. Formulations has been adjusted with interval ``config.t``
    2. The tie-line flow has not been implemented in formulations.

    References
    ----------
    1. Huang, Y., Pardalos, P. M., & Zheng, Q. P. (2017). Electrical power unit commitment: deterministic and
       two-stage stochastic programming models and algorithms. Springer.
    2. D. A. Tejada-Arango, S. Lumbreras, P. Sánchez-Martín and A. Ramos, "Which Unit-Commitment Formulation
       is Best? A Comparison Framework," in IEEE Transactions on Power Systems, vol. 35, no. 4, pp. 2926-2936,
       July 2020, doi: 10.1109/TPWRS.2019.2962024.
    """

    def __init__(self, system, config):
        DCOPF.__init__(self, system, config)
        RTEDBase.__init__(self)
        MPBase.__init__(self)
        SRBase.__init__(self)
        NSRBase.__init__(self)

        self.config.add(OrderedDict((('t', 1),
                                     )))
        self.config.add_extra("_help",
                              t="time interval in hours",
                              )
        self.config.add_extra("_tex",
                              t='T_{cfg}',
                              )

        self.info = 'unit commitment'
        self.type = 'DCUC'

        # --- Data Section ---
        # update timeslot model to UCTSlot
        self.timeslot.info = 'Time slot for multi-period UC'
        self.timeslot.model = 'UCTSlot'
        # --- reserve cost ---
        self.csu = RParam(info='startup cost',
                          name='csu', tex_name=r'c_{su}',
                          model='GCost', src='csu',
                          unit='$',)
        self.csd = RParam(info='shutdown cost',
                          name='csd', tex_name=r'c_{sd}',
                          model='GCost', src='csd',
                          unit='$',)
        # --- load ---
        self.cdp = RParam(info='penalty for unserved load',
                          name='cdp', tex_name=r'c_{d,p}',
                          model='DCost', src='cdp',
                          no_parse=True,
                          unit=r'$/(p.u.*h)',)
        self.dctrl = RParam(info='load controllability',
                            name='dctrl', tex_name=r'c_{trl,d}',
                            model='StaticLoad', src='ctrl',
                            expand_dims=1,
                            no_parse=True,)
        # --- gen ---
        self.td1 = RParam(info='minimum ON duration',
                          name='td1', tex_name=r't_{d1}',
                          model='StaticGen', src='td1',
                          unit='h',)
        self.td2 = RParam(info='minimum OFF duration',
                          name='td2', tex_name=r't_{d2}',
                          model='StaticGen', src='td2',
                          unit='h',)

        self.sd.info = 'zonal load factor for UC'
        self.sd.model = 'UCTSlot'

        self.ug.expand_dims = 1
        self.amin.expand_dims = 1
        self.amax.expand_dims = 1

        # --- Model Section ---
        # --- gen ---
        # NOTE: extend pg to 2D matrix, where row for gen and col for timeslot
        self.pg.horizon = self.timeslot
        self.pg.info = '2D Gen power'
        # TODO: havn't test non-controllability?
        self.ctrle.u2 = self.tlv
        self.ctrle.info = 'Reshaped controllability'
        self.nctrle.u2 = self.tlv
        self.nctrle.info = 'Reshaped non-controllability'
        pmaxe = 'mul(mul(nctrl, pg0), ugd) + mul(mul(ctrl, pmax), ugd)'
        self.pmaxe.e_str = pmaxe
        self.pmaxe.horizon = self.timeslot
        pmine = 'mul(mul(ctrl, pmin), ugd) + mul(mul(nctrl, pg0), ugd)'
        self.pmine.e_str = pmine
        self.pmine.horizon = self.timeslot
        self.pglb.e_str = '-pg + pmine'
        self.pgub.e_str = 'pg - pmaxe'

        self.ugd = Var(info='commitment decision',
                       horizon=self.timeslot,
                       name='ugd', tex_name=r'u_{g,d}',
                       model='StaticGen', src='u',
                       boolean=True,)
        self.vgd = Var(info='startup action',
                       horizon=self.timeslot,
                       name='vgd', tex_name=r'v_{g,d}',
                       model='StaticGen', src='u',
                       boolean=True,)
        self.wgd = Var(info='shutdown action',
                       horizon=self.timeslot,
                       name='wgd', tex_name=r'w_{g,d}',
                       model='StaticGen', src='u',
                       boolean=True,)
        self.zug = Var(info='Aux var, :math:`z_{ug} = u_{g,d} * p_g`',
                       horizon=self.timeslot,
                       name='zug', tex_name=r'z_{ug}',
                       model='StaticGen', pos=True,)
        # NOTE: actions have two parts: initial status and the rest
        self.actv = Constraint(name='actv', is_eq=True,
                               info='startup action',
                               e_str='ugd @ Mr - vgd[:, 1:]',)
        self.actv0 = Constraint(name='actv0', is_eq=True,
                                info='initial startup action',
                                e_str='ugd[:, 0] - ug[:, 0]  - vgd[:, 0]',)
        self.actw = Constraint(name='actw', is_eq=True,
                               info='shutdown action',
                               e_str='-ugd @ Mr - wgd[:, 1:]',)
        self.actw0 = Constraint(name='actw0', is_eq=True,
                                info='initial shutdown action',
                                e_str='-ugd[:, 0] + ug[:, 0] - wgd[:, 0]',)

        self.prs.horizon = self.timeslot
        self.prs.info = '2D Spinning reserve'

        self.prns.horizon = self.timeslot
        self.prns.info = '2D Non-spinning reserve'

        # spinning reserve
        self.prsb.e_str = 'mul(ugd, mul(pmax, tlv)) - zug - prs'
        # spinning reserve requirement
        self.rsr.e_str = '-gs@prs + dsr'

        # non-spinning reserve
        self.prnsb.e_str = 'mul(1-ugd, mul(pmax, tlv)) - prns'
        # non-spinning reserve requirement
        self.rnsr.e_str = '-gs@prns + dnsr'

        # --- big M for ugd*pg ---
        self.Mzug = NumOp(info='10 times of max of pmax as big M for zug',
                          name='Mzug', tex_name=r'M_{zug}',
                          u=self.pmax, fun=np.max,
                          rfun=np.dot, rargs=dict(b=10),
                          array_out=False,)
        self.zuglb = Constraint(name='zuglb', info='zug lower bound',
                                is_eq=False, e_str='- zug + pg')
        self.zugub = Constraint(name='zugub', info='zug upper bound',
                                is_eq=False, e_str='zug - pg - Mzug dot (1 - ugd)')
        self.zugub2 = Constraint(name='zugub2', info='zug upper bound',
                                 is_eq=False, e_str='zug - Mzug dot ugd')

        # --- minimum ON/OFF duration ---
        self.Con = MinDur(u=self.pg, u2=self.td1,
                          name='Con', tex_name=r'T_{on}',
                          info='minimum ON coefficient',)
        self.don = Constraint(info='minimum online duration',
                              name='don', is_eq=False,
                              e_str='multiply(Con, vgd) - ugd')
        self.Coff = MinDur(u=self.pg, u2=self.td2,
                           name='Coff', tex_name=r'T_{off}',
                           info='minimum OFF coefficient',)
        self.doff = Constraint(info='minimum offline duration',
                               name='doff', is_eq=False,
                               e_str='multiply(Coff, wgd) - (1 - ugd)')

        # --- line ---
        self.plf.horizon = self.timeslot
        self.plf.info = '2D Line flow'
        self.plflb.e_str = '-Bf@aBus - Pfinj - mul(rate_a, tlv)'
        self.plfub.e_str = 'Bf@aBus + Pfinj - mul(rate_a, tlv)'
        self.alflb.e_str = '-CftT@aBus - amax@tlv'
        self.alfub.e_str = 'CftT@aBus - amax@tlv'

        # --- unserved load ---
        self.pdu = Var(info='unserved demand',
                       name='pdu', tex_name=r'p_{d,u}',
                       model='StaticLoad', unit='p.u.',
                       horizon=self.timeslot,
                       nonneg=True,)
        self.pdsp = NumOp(u=self.pds, fun=np.clip,
                          args=dict(a_min=0, a_max=None),
                          info='positive demand',
                          name='pdsp', tex_name=r'p_{d,s}^{+}',)
        self.pdumax = Constraint(info='unserved demand upper bound',
                                 name='pdumax', is_eq=False,
                                 e_str='pdu - mul(pdsp, dctrl@tlv)')
        # --- power balance ---
        # NOTE: nodal balance is also contributed by unserved load
        pb = 'Bbus@aBus + Pbusinj@tlv + Cl@(pds-pdu) + Csh@gsh@tlv - Cg@pg'
        self.pb.e_str = pb

        # --- objective ---
        cost = 't**2 dot sum(c2 @ zug**2 + t dot c1 @ zug)'
        cost += '+ sum(mul(ug, c0) @ tlv)'
        _to_sum = 'csu * vgd + csd * wgd + csr @ prs + cnsr @ prns + cdp @ pdu'
        cost += f' + t dot sum({_to_sum})'
        self.obj.e_str = cost

    def _initial_guess(self):
        """
        Make initial guess for commitment decision with a priority list
        defined by the weighted sum of generation cost and generator capacity.

        If there are no offline generators, turn off the first 30% of the generators
        on the priority list as initial guess.
        """
        # check trigger condition
        ug0 = self.system.PV.get(src='u', attr='v', idx=self.system.PV.idx.v)
        if (ug0 == 0).any():
            return True
        else:
            logger.warning('All generators are online at initial, make initial guess for commitment.')

        gen = pd.DataFrame()
        gen['idx'] = self.system.PV.idx.v
        gen['pmax'] = self.system.PV.get(src='pmax', attr='v', idx=gen['idx'])
        gen['bus'] = self.system.PV.get(src='bus', attr='v', idx=gen['idx'])
        gen['area'] = self.system.PV.get(src='area', attr='v', idx=gen['idx'])
        gcost_idx = self.system.GCost.find_idx(keys='gen', values=gen['idx'])
        gen['c2'] = self.system.GCost.get(src='c2', attr='v', idx=gcost_idx)
        gen['c1'] = self.system.GCost.get(src='c1', attr='v', idx=gcost_idx)
        gen['c0'] = self.system.GCost.get(src='c0', attr='v', idx=gcost_idx)
        gen['wsum'] = 0.8*gen['pmax'] + 0.05*gen['c2'] + 0.1*gen['c1'] + 0.05*gen['c0']
        gen = gen.sort_values(by='wsum', ascending=True)

        # Turn off 30% of the generators as initial guess
        priority = gen['idx'].values
        g_idx = priority[0:int(0.3*len(priority))]
        ug0 = np.zeros_like(g_idx)
        # NOTE: if number of generators is too small, turn off the first one
        if len(g_idx) == 0:
            g_idx = priority[0]
            ug0 = 0
            off_gen = f'{g_idx}'
        else:
            off_gen = ', '.join(g_idx)
        self.system.StaticGen.set(src='u', idx=g_idx, attr='v', value=ug0)
        logger.warning(f"As initial commitment guess, turn off StaticGen: {off_gen}")
        return g_idx

    def init(self, **kwargs):
        self._initial_guess()
        return super().init(**kwargs)

    def dc2ac(self, kloss=1.0, **kwargs):
        """
        AC conversion ``dc2ac`` is not implemented yet for
        multi-period scheduling.
        """
        return NotImplementedError

    def unpack(self, res, **kwargs):
        """
        Multi-period scheduling will not unpack results from
        solver into devices.

        # TODO: unpack first period results, and allow input
        # to specify which period to unpack.
        """
        return None


class UCDG(UC, DGBase):
    """
    UC with distributed generation :ref:`DG`.

    Note that UCDG only inlcudes DG output power. If ESD1 is included,
    UCES should be used instead, otherwise there is no SOC.
    """

    def __init__(self, system, config):
        UC.__init__(self, system, config)
        DGBase.__init__(self)

        self.info = 'unit commitment with distributed generation'
        self.type = 'DCUC'

        # NOTE: extend vars to 2D
        self.pgdg.horizon = self.timeslot


class UCES(UC, ESD1MPBase):
    """
    UC with energy storage :ref:`ESD1`.
    """

    def __init__(self, system, config):
        UC.__init__(self, system, config)
        ESD1MPBase.__init__(self)

        self.info = 'unit commitment with energy storage'
        self.type = 'DCUC'

        self.pgdg.horizon = self.timeslot
        self.SOC.horizon = self.timeslot
        self.pce.horizon = self.timeslot
        self.pde.horizon = self.timeslot
        self.uce.horizon = self.timeslot
        self.ude.horizon = self.timeslot
        self.zce.horizon = self.timeslot
        self.zde.horizon = self.timeslot



from collections import OrderedDict

from andes.core.param import NumParam, ExtParam, DataParam
from andes.models.static.pv import PVData
from andes.models.static.slack import SlackData

from ams.core.model import Model
from ams.core.var import Algeb


class GenParam:
    """
    Additional parameters for static generators.
    """

    def __init__(self) -> None:
        self.ctrl = NumParam(default=1,
                             info="generator controllability",
                             tex_name=r'c_{trl}',
                             unit='bool',)
        self.uf = NumParam(default=0,
                           info="enforced on/off status; 0 unenforced, 1 enforced on, -1 enforced off",
                           tex_name=r'u_{f}',
                           unit='int',
                           vrange=(-1, 1))
        self.Pc1 = NumParam(default=0.0,
                            info="lower real power output of PQ capability curve",
                            tex_name=r'P_{c1}',
                            unit='p.u.')
        self.Pc2 = NumParam(default=0.0,
                            info="upper real power output of PQ capability curve",
                            tex_name=r'P_{c2}',
                            unit='p.u.')
        self.Qc1min = NumParam(default=0.0,
                               info="minimum reactive power output at Pc1",
                               tex_name=r'Q_{c1min}',
                               unit='p.u.')
        self.Qc1max = NumParam(default=0.0,
                               info="maximum reactive power output at Pc1",
                               tex_name=r'Q_{c1max}',
                               unit='p.u.')
        self.Qc2min = NumParam(default=0.0,
                               info="minimum reactive power output at Pc2",
                               tex_name=r'Q_{c2min}',
                               unit='p.u.')
        self.Qc2max = NumParam(default=0.0,
                               info="maximum reactive power output at Pc2",
                               tex_name=r'Q_{c2max}',
                               unit='p.u.')
        self.Ragc = NumParam(default=999.0,
                             info="ramp rate for load following/AGC",
                             tex_name=r'R_{agc}',
                             unit='p.u./h')
        self.R10 = NumParam(default=999.0,
                            info="ramp rate for 10 minute reserves",
                            tex_name=r'R_{10}',
                            unit='p.u./h')
        self.R30 = NumParam(default=999.0,
                            info="30 minute ramp rate",
                            tex_name=r'R_{30}',
                            unit='p.u./h')
        self.Rq = NumParam(default=999.0,
                           info="ramp rate for reactive power (2 sec timescale)",
                           tex_name=r'R_{q}',
                           unit='p.u./h')
        self.apf = NumParam(default=0.0,
                            info="area participation factor",
                            tex_name=r'a_{pf}')
        self.pg0 = NumParam(default=0.0,
                            info='active power start point (system base)',
                            tex_name=r'p_{g0}',
                            unit='p.u.',
                            )
        self.td1 = NumParam(default=1,
                            info='minimum ON duration',
                            tex_name=r't_{d1}',
                            unit='h',
                            )
        self.td2 = NumParam(default=0.5,
                            info='minimum OFF duration',
                            tex_name=r't_{d2}',
                            unit='h',
                            )
        self.gentype = DataParam(default=None,
                                 info='generator type',
                                 tex_name=r'g_{type}',
                                 )
        self.genfuel = DataParam(default=None,
                                 info='generator fuel type',
                                 tex_name=r'g_{fuel}',
                                 )


class PVModel(Model):
    """
    PV generator model (power flow) with q limit and PV-PQ conversion.
    """

    def __init__(self, system=None, config=None):
        super().__init__(system, config)
        self.group = 'StaticGen'

        self.config.add(OrderedDict((('pv2pq', 0),
                                     ('npv2pq', 0),
                                     ('min_iter', 2),
                                     ('err_tol', 0.01),
                                     ('abs_violation', 1),
                                     )))
        self.config.add_extra("_help",
                              pv2pq="convert PV to PQ in PFlow at Q limits",
                              npv2pq="max. # of conversion each iteration, 0 - auto",
                              min_iter="iteration number starting from which to enable switching",
                              err_tol="iteration error below which to enable switching",
                              abs_violation='use absolute (1) or relative (0) limit violation',
                              )

        self.config.add_extra("_alt",
                              pv2pq=(0, 1),
                              npv2pq=">=0",
                              min_iter='int',
                              err_tol='float',
                              abs_violation=(0, 1),
                              )
        self.config.add_extra("_tex",
                              pv2pq="z_{pv2pq}",
                              npv2pq="n_{pv2pq}",
                              min_iter="sw_{iter}",
                              err_tol=r"\epsilon_{tol}"
                              )

        self.area = ExtParam(model='Bus', src='area', indexer=self.bus, export=False,
                             info='Retrieved area idx', vtype=str, default=None,
                             )

        self.ud = Algeb(info='commitment decision',
                        unit='bool',
                        tex_name=r'u_d',
                        name='ud',
                        )
        self.p = Algeb(info='active power generation',
                       unit='p.u.',
                       tex_name='p',
                       name='p',
                       )
        self.q = Algeb(info='reactive power generation',
                       unit='p.u.',
                       tex_name='q',
                       name='q',
                       )


class PV(PVData, GenParam, PVModel):
    """
    PV generator model.

    TODO: implement type conversion in config
    """

    def __init__(self, system, config):
        PVData.__init__(self)
        GenParam.__init__(self)
        PVModel.__init__(self, system, config)


class Slack(SlackData, GenParam, PVModel):
    """
    Slack generator model.
    """

    def __init__(self, system=None, config=None):
        SlackData.__init__(self)
        GenParam.__init__(self)
        PVModel.__init__(self, system, config)



import contextlib
import unittest

import ams


class TestRepr(unittest.TestCase):
    """Test __repr__"""
    def setUp(self):
        self.ss = ams.run(ams.get_case("ieee39/ieee39_uced.xlsx"),
                          no_output=True,
                          default_config=True,
                          )

    def test_print_model_repr(self):
        """
        Print out Model ``cache``'s fields and values.
        """
        with contextlib.redirect_stdout(None):
            for model in self.ss.models.values():
                print(model.cache.__dict__)



from andes.models.line.line import LineData
from andes.models.line.jumper import JumperData
from andes.core.param import NumParam
from andes.shared import deg2rad

from ams.core.model import Model


class Line(LineData, Model):
    """
    AC transmission line model.

    The model is also used for two-winding transformer. Transformers can set the
    tap ratio in ``tap`` and/or phase shift angle ``phi``.

    Note that the bus admittance matrix is built on fly and is not stored in the
    object.

    Notes
    -----
    1. Adding Algeb ``ud`` causes Line.algebs to encounter an AttributeError: 'NoneType'
       object has no attribute 'n'. The root cause is still under investigation.
    """

    def __init__(self, system=None, config=None) -> None:
        LineData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'ACLine'

        self.amin = NumParam(default=-360 * deg2rad,
                             info="minimum angle difference, from bus - to bus",
                             unit='rad',
                             tex_name=r'a_{min}',
                             )
        self.amax = NumParam(default=360 * deg2rad,
                             info="maximum angle difference, from bus - to bus",
                             unit='rad',
                             tex_name=r'a_{max}',
                             )
        self.rate_a.unit = 'p.u.'
        self.rate_b.unit = 'p.u.'
        self.rate_c.unit = 'p.u.'
        self.rate_a.default = 999.0
        self.rate_b.default = 999.0
        self.rate_c.default = 999.0

        # NOTE: following parameters are prepared for building matrices
        # they are initialized here but populated in ``System.setup()``.
        self.a1a = None
        self.a2a = None


class Jumper(JumperData, Model):
    """
    Jumper is a device to short two buses (merging two buses into one).

    Jumper can connect two buses satisfying one of the following conditions:

    - neither bus is voltage-controlled
    - either bus is voltage-controlled
    - both buses are voltage-controlled, and the voltages are the same.

    If the buses are controlled in different voltages, power flow will
    not solve (as the power flow through the jumper will be infinite).

    In the solutions, the ``p`` and ``q`` are flowing out of bus1
    and flowing into bus2.

    Setting a Jumper's connectivity status ``u`` to zero will disconnect the two
    buses. In the case of a system split, one will need to call
    ``System.connectivity()`` immediately following the split to detect islands.
    """

    def __init__(self, system=None, config=None) -> None:
        JumperData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'ACShort'



"""
RenGen scheduling model.
"""

from andes.core.param import NumParam, IdxParam
from andes.core.model import ModelData
from ams.core.model import Model


class REGCData(ModelData):
    """
    Data container for RenGen scheduling model.
    """

    def __init__(self):
        ModelData.__init__(self)
        self.bus = IdxParam(model='Bus',
                            info="interface bus idx",
                            mandatory=True,
                            )
        self.gen = IdxParam(info="static generator index",
                            mandatory=True,
                            )
        self.Sn = NumParam(default=100.0, tex_name='S_n',
                           info='device MVA rating',
                           unit='MVA',
                           )
        self.gammap = NumParam(default=1.0,
                               info="P ratio of linked static gen",
                               tex_name=r'\gamma_P'
                               )
        self.gammaq = NumParam(default=1.0,
                               info="Q ratio of linked static gen",
                               tex_name=r'\gamma_Q'
                               )


class REGCA1(REGCData, Model):
    """
    Renewable generator scheduling model.

    Reference:

    [1] ANDES Documentation, REGCA1

    Available:
    https://docs.andes.app/en/latest/groupdoc/RenGen.html#regca1
    """

    def __init__(self, system=None, config=None) -> None:
        REGCData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'RenGen'


class REGCV1(REGCData, Model):
    """
    Voltage-controlled converter model (virtual synchronous generator) with
    inertia emulation.

    Here Mmax and Dmax are assumed to be constant, but they might subject to
    the operating condition of the converter.

    Notes
    -----
    - The generation is defined by group :ref:`StaticGen`
    - Generation cost is defined by model :ref:`GCost`
    - Inertia emulation cost is defined by model :ref:`VSGCost`

    Reference:

    [1] ANDES Documentation, REGCV1

    Available:

    https://docs.andes.app/en/latest/groupdoc/RenGen.html#regcv1
    """

    def __init__(self, system=None, config=None) -> None:
        REGCData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'VSG'
        self.M = NumParam(default=10, tex_name='M',
                          info='Inertia emulation',
                          unit='s',
                          power=True,)
        self.D = NumParam(default=0, tex_name='D',
                          info='Damping emulation',
                          unit='p.u.',
                          power=True,)
        self.Mmax = NumParam(default=99, tex_name='M_{max}',
                             info='Maximum inertia emulation',
                             unit='s',
                             power=True,)
        self.Dmax = NumParam(default=99, tex_name='D_{max}',
                             info='Maximum damping emulation',
                             unit='p.u.',
                             power=True,)


class REGCV2(REGCV1):
    """
    Voltage-controlled VSC, identical to :ref:`REGCV1`.

    Reference:

    [1] ANDES Documentation, REGCV2

    Available:

    https://docs.andes.app/en/latest/groupdoc/RenGen.html#regcv2
    """

    def __init__(self, system=None, config=None) -> None:
        REGCV1.__init__(self, system, config)



"""
Module for optimization modeling.
"""

from ams.opt.optzbase import OptzBase, ensure_symbols, ensure_mats_and_parsed  # NOQA
from ams.opt.var import Var  # NOQA
from ams.opt.exprcalc import ExpressionCalc  # NOQA
from ams.opt.param import Param  # NOQA
from ams.opt.constraint import Constraint  # NOQA
from ams.opt.objective import Objective  # NOQA
from ams.opt.expression import Expression  # NOQA
from ams.opt.omodel import OModelBase, OModel  # NOQA



"""
Test wrapper routines of PYPOWER.
"""
import unittest

import ams
from ams.shared import skip_unittest_without_PYPOWER


class TestDCPF1(unittest.TestCase):
    """
    Test routine `DCPF1`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.json"),
                           setup=True, default_config=True, no_output=True)
        # decrease load first
        self.ss.PQ.set(src='p0', attr='v', idx=['PQ_1', 'PQ_2'], value=[0.3, 0.3])

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.DCPF1.init()
        self.assertTrue(self.ss.DCPF1.initialized, "DCPF1 initialization failed!")

    @skip_unittest_without_PYPOWER
    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        stg = 'PV_1'
        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=0)

        self.ss.DCPF1.update()
        self.ss.DCPF1.run()
        self.assertTrue(self.ss.DCPF1.converged, "DCPF1 did not converge under generator trip!")
        self.assertAlmostEqual(self.ss.DCPF1.get(src='pg', attr='v', idx=stg),
                               0, places=6,
                               msg="Generator trip does not take effect!")

        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=1)  # reset

    @skip_unittest_without_PYPOWER
    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx='Line_3', value=0)

        self.ss.DCPF1.update()
        self.ss.DCPF1.run()
        self.assertTrue(self.ss.DCPF1.converged, "DCPF1 did not converge under line trip!")
        self.assertAlmostEqual(self.ss.DCPF1.get(src='plf', attr='v', idx='Line_3'),
                               0, places=6,
                               msg="Line trip does not take effect!")

        self.ss.Line.alter(src='u', idx='Line_3', value=1)  # reset

    @skip_unittest_without_PYPOWER
    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        # --- run DCPF1 ---
        self.ss.DCPF1.run()
        pgs = self.ss.DCPF1.pg.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.DCPF1.update()

        self.ss.DCPF1.run()
        pgs_pqt = self.ss.DCPF1.pg.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.set(src='u', attr='v', idx='PQ_2', value=0)
        self.ss.DCPF1.update()

        self.ss.DCPF1.run()
        pgs_pqt2 = self.ss.DCPF1.pg.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")


class TestPFlow1(unittest.TestCase):
    """
    Test routine `PFlow1`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.json"),
                           setup=True, default_config=True, no_output=True)
        # decrease load first
        self.ss.PQ.set(src='p0', attr='v', idx=['PQ_1', 'PQ_2'], value=[0.3, 0.3])

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.PFlow1.init()
        self.assertTrue(self.ss.PFlow1.initialized, "PFlow1 initialization failed!")

    @skip_unittest_without_PYPOWER
    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        stg = 'PV_1'
        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=0)

        self.ss.PFlow1.update()
        self.ss.PFlow1.run()
        self.assertTrue(self.ss.PFlow1.converged, "PFlow1 did not converge under generator trip!")
        self.assertAlmostEqual(self.ss.PFlow1.get(src='pg', attr='v', idx=stg),
                               0, places=6,
                               msg="Generator trip does not take effect!")

        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=1)  # reset

    @skip_unittest_without_PYPOWER
    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx='Line_3', value=0)

        self.ss.PFlow1.update()
        self.ss.PFlow1.run()
        self.assertTrue(self.ss.PFlow1.converged, "PFlow1 did not converge under line trip!")
        self.assertAlmostEqual(self.ss.PFlow1.get(src='plf', attr='v', idx='Line_3'),
                               0, places=6,
                               msg="Line trip does not take effect!")

        self.ss.Line.alter(src='u', idx='Line_3', value=1)  # reset

    @skip_unittest_without_PYPOWER
    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        # --- run PFlow1 ---
        self.ss.PFlow1.run()
        pgs = self.ss.PFlow1.pg.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.PFlow1.update()

        self.ss.PFlow1.run()
        pgs_pqt = self.ss.PFlow1.pg.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.set(src='u', attr='v', idx='PQ_2', value=0)
        self.ss.PFlow1.update()

        self.ss.PFlow1.run()
        pgs_pqt2 = self.ss.PFlow1.pg.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")


class TestDCOPF1(unittest.TestCase):
    """
    Test routine `DCOPF1`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.json"),
                           setup=True, default_config=True, no_output=True)
        # decrease load first
        self.ss.PQ.set(src='p0', attr='v', idx=['PQ_1', 'PQ_2'], value=[0.3, 0.3])

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.DCOPF1.init()
        self.assertTrue(self.ss.DCOPF1.initialized, "DCOPF initialization failed!")

    @skip_unittest_without_PYPOWER
    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        stg = 'PV_1'
        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=0)

        self.ss.DCOPF1.update()
        self.ss.DCOPF1.run()
        self.assertTrue(self.ss.DCOPF1.converged, "DCOPF did not converge under generator trip!")
        self.assertAlmostEqual(self.ss.DCOPF1.get(src='pg', attr='v', idx=stg),
                               0, places=6,
                               msg="Generator trip does not take effect!")

        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=1)  # reset

    @skip_unittest_without_PYPOWER
    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx='Line_3', value=0)

        self.ss.DCOPF1.update()
        self.ss.DCOPF1.run()
        self.assertTrue(self.ss.DCOPF1.converged, "DCOPF1 did not converge under line trip!")
        self.assertAlmostEqual(self.ss.DCOPF1.get(src='plf', attr='v', idx='Line_3'),
                               0, places=6,
                               msg="Line trip does not take effect!")

        self.ss.Line.alter(src='u', idx='Line_3', value=1)  # reset

    @skip_unittest_without_PYPOWER
    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        # --- run DCOPF1 ---
        self.ss.DCOPF1.run()
        pgs = self.ss.DCOPF1.pg.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.DCOPF1.update()

        self.ss.DCOPF1.run()
        pgs_pqt = self.ss.DCOPF1.pg.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.set(src='u', attr='v', idx='PQ_2', value=0)
        self.ss.DCOPF1.update()

        self.ss.DCOPF1.run()
        pgs_pqt2 = self.ss.DCOPF1.pg.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")


class TestACOPF1(unittest.TestCase):
    """
    Test routine `ACOPF1`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.json"),
                           setup=True, default_config=True, no_output=True)
        # decrease load first
        self.ss.PQ.set(src='p0', attr='v', idx=['PQ_1', 'PQ_2'], value=[0.3, 0.3])

    def test_init(self):
        """
        Test initialization.
        """
        self.ss.ACOPF1.init()
        self.assertTrue(self.ss.ACOPF1.initialized, "ACOPF1 initialization failed!")

    @skip_unittest_without_PYPOWER
    def test_trip_gen(self):
        """
        Test generator tripping.
        """
        stg = 'PV_1'
        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=0)

        self.ss.ACOPF1.update()
        self.ss.ACOPF1.run()
        self.assertTrue(self.ss.ACOPF1.converged, "ACOPF1 did not converge under generator trip!")
        self.assertAlmostEqual(self.ss.ACOPF1.get(src='pg', attr='v', idx=stg),
                               0, places=6,
                               msg="Generator trip does not take effect!")

        self.ss.StaticGen.set(src='u', idx=stg, attr='v', value=1)  # reset

    @skip_unittest_without_PYPOWER
    def test_trip_line(self):
        """
        Test line tripping.
        """
        self.ss.Line.set(src='u', attr='v', idx='Line_3', value=0)

        self.ss.ACOPF1.update()
        self.ss.ACOPF1.run()
        self.assertTrue(self.ss.ACOPF1.converged, "ACOPF1 did not converge under line trip!")
        self.assertAlmostEqual(self.ss.ACOPF1.get(src='plf', attr='v', idx='Line_3'),
                               0, places=6,
                               msg="Line trip does not take effect!")

        self.ss.Line.alter(src='u', idx='Line_3', value=1)  # reset

    @skip_unittest_without_PYPOWER
    def test_set_load(self):
        """
        Test setting and tripping load.
        """
        # --- run ACOPF1 ---
        self.ss.ACOPF1.run()
        pgs = self.ss.ACOPF1.pg.v.sum()

        # --- set load ---
        self.ss.PQ.set(src='p0', attr='v', idx='PQ_1', value=0.1)
        self.ss.ACOPF1.update()

        self.ss.ACOPF1.run()
        pgs_pqt = self.ss.ACOPF1.pg.v.sum()
        self.assertLess(pgs_pqt, pgs, "Load set does not take effect!")

        # --- trip load ---
        self.ss.PQ.set(src='u', attr='v', idx='PQ_2', value=0)
        self.ss.ACOPF1.update()

        self.ss.ACOPF1.run()
        pgs_pqt2 = self.ss.ACOPF1.pg.v.sum()
        self.assertLess(pgs_pqt2, pgs_pqt, "Load trip does not take effect!")



"""
Reserve model.
"""

from andes.core import (ModelData, IdxParam, NumParam)

from ams.core.model import Model


class ReserveData(ModelData):
    def __init__(self):
        super().__init__()
        self.area = IdxParam(model='Area',
                             default=None, info="Area idx",)


class SFR(ReserveData, Model):
    """
    Zonal secondary frequency reserve (SFR) model.

    Notes
    -----
    - ``Zone`` model is required for this model, and zone is defined by Param ``Bus.zone``.
    """

    def __init__(self, system, config):
        ReserveData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Reserve'
        self.du = NumParam(default=0,
                           info='Zonal RegUp reserve demand',
                           tex_name=r'd_{u}',)
        self.dd = NumParam(default=0,
                           info='Zonal RegDown reserve demand',
                           tex_name=r'd_{d}',)


class SR(ReserveData, Model):
    """
    Zonal spinning reserve (SR) model.

    Notes
    -----
    - ``Zone`` model is required for this model, and zone is defined by Param ``Bus.zone``.
    - ``demand`` is multiplied to online unused generation capacity.
    """

    def __init__(self, system, config):
        ReserveData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Reserve'
        self.demand = NumParam(default=0.1,
                               info='Zonal spinning reserve demand',
                               tex_name=r'd_{SR}')


class NSR(ReserveData, Model):
    """
    Zonal non-spinning reserve (NSR) model.

    Notes
    -----
    - ``Zone`` model is required for this model, and zone is defined by Param ``Bus.zone``.
    - ``demand`` is multiplied to offline generation capacity.
    """

    def __init__(self, system, config):
        ReserveData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Reserve'
        self.demand = NumParam(default=0.1,
                               info='Zonal non-spinning reserve demand',
                               tex_name=r'd_{NSR}')


class VSGR(ReserveData, Model):
    """
    Zonal VSG provided reserve model.

    Notes
    -----
    - ``Zone`` model is required for this model, and zone is defined by Param ``Bus.zone``.
    """

    def __init__(self, system, config):
        ReserveData.__init__(self)
        Model.__init__(self, system, config)
        self.group = 'Reserve'
        self.dvm = NumParam(default=0, unit='s',
                            info='Zonal virtual inertia demand',
                            tex_name=r'd_{VM}')
        self.dvd = NumParam(default=0, unit='p.u.',
                            info='Zonal virtual damping demand',
                            tex_name=r'd_{VD}')
